#!/usr/bin/env bash
#
# Simple Bash script to help build a binary distribution archive for various
# Uni-Verse programs. Also includes source, just because we can.
#

ROOT=../..
DATE=$(date +"%Y%m%d")
PLATFORM=Linux
PLTFRM=$(echo "$PLATFORM" | tr [A-Z] [a-z] | tr -d ' ')
OUT=universe-$PLTFRM-$DATE

# Emit a line of documentation.
function doc {
	echo $* >>$OUT/index.html
}

# Emit documentation header, with CVS link.
function doch {
	doc "<h2>$1</h2> [CVS: <a href=\"http://projects.blender.org/viewcvs/viewcvs.cgi/$2/?cvsroot=verse\"><tt>$2</tt></a>]"
}

function doc-header {
	echo "<html>" >$OUT/index.html
	doc "<head>"
	doc "<title>Uni-Verse Software Release 2," $PLATFORM "</title>"
	doc '<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">'
	doc '<style type="text/css">'
	doc ' h1.title { text-align: center; }'
	doc ' p.subtitle { text-align: center; }'
	doc ' h2 { background-color: #eeeeff; margin-bottom: 0;}'
	doc '</style>'
	doc "</head><body>"
	doc '<h1 class="title">Uni-Verse Software Release 2</h1>'
	doc '<p class="subtitle">' $PLATFORM ' Version, compiled on ' $DATE '</p></center>'

	doc '<h1>Introduction</h1>'
	doc '<p>'
	doc 'Welcome! This is the second release of <a href="http://www.uni-verse.org/">Uni-Verse</a> tools. We hope you enjoy the software.'
	doc '</p>'
	doc '<h1>Feedback</h1>'
	doc '<p>'
	doc 'Feedback to the developers of this software is very much appreciated. The best way is by e-mail to the'
	doc '<a href="http://projects.blender.org/mailman/listinfo/universe-feedback">uni-verse feedback</a> mailing list. Please'
	doc 'subscribe yourself to the list before trying to post to it, and also consider looking for archived posts. Thank you.'
	doc '</p>'
	doc '<h1>Software</h1>'
	doc '<p>'
	doc 'Below are the various pieces of software included in this release. The links go to straight to the application binaries,'
	doc 'which generally does not work to start applications in Linux. You would do better to just open a terminal window, cd to'
	doc 'wherever this HTML file is, and play with the programs directly.'
	doc '</p>'
	doc '<p>'
	doc 'The CVS links below go straight into the Verse CVS repository, where the software source code lives. You are heartily '
	doc 'recommended to read the <a href="http://projects.blender.org/cvsx/?group_id=35">instructions</a> and check out any module '
	doc 'you are interested in yourself. Do please note that for the core <code>verse</code> CVS module, you need to specify a tagged '
	doc 'release, namely <code>release-r6-branch</code>. For all other CVS modules, use HEAD as normal.'
	doc '</p>'
	doc '<p>'
	doc 'To run the binaries included here, you will need several libraries on your machine. We do not include these libraries in '
	doc 'the distribution, but instead rely on the various Linux distributions to provide them. The required libraries include: '
	doc '</p>'
	doc '<ul>'
	doc '<li><a href="http://www.opengl.org/">OpenGL</a>, with GLUT and GLU'
	doc '<li><a href="http://www.libsdl.org/">SDL</a>'
	doc '</ul>'
	doc '<p>'
	doc 'To run Puyo, you will also need <a href="http://www.python.org/">Python</a> and quite a number of supporting Python modules.'
	doc 'See the relevant chapter of the <a href="docs/puyo/inst-deps.html">included documentation</a> for details.'
	doc '</p>'
}

function doc-footer {
	doc "</body></html>"
}

function dist {
	for a in $1
	do
		cp $ROOT/$a $OUT
	done

	# Any documentation specified? Assumed to be directories
	# to be copied recursively into $OUT/docs.
	if [[ $# == 2 ]]; then
		echo "copying docs $2"
		for d in $2
		do
			cp -r $ROOT/$d $OUT/docs
		done
	fi
}

function dist-verse {
	( cd $ROOT/verse && make )
	FILES="verse/libverse.a verse/verse.h verse/verse"
	dist "$FILES"

	doch 'Verse Core' 'verse'
	doc '<p>'
	doc 'This is the core Verse software. The only directly runnable part here is the server, which is typically required'
	doc 'for all the other applications to work. The other two files are for people who wish to develop their own Verse'
	doc 'applications.'
	doc '</p>'
	doc '<p><dl>'
	doc '<dt><a href="verse">verse server</a></dt><dd>The Verse server. Start this before trying to run any of the other applications.</dd>'
	doc '<dt><a href="verse.h" type="text/plain">verse.h</a></dt><dd>This is a header file that defines the Verse API for C programmers.'
	doc 'Goes well with the <a href="libverse.a">libverse.a</a> link library also found here.</dd>'
	doc '</p></dl>'
}

function dist-simple {
	( cd $ROOT/verse-tests && make render )
	FILES="verse-tests/render"
	DOCS="verse-tests/docs/render"
	dist "$FILES" "$DOCS"

	doch 'Simple Renderer' 'verse-tests'
	doc '<p>'
	doc 'This is a simple rendering client. It does not offer high-quality rendering, but is small and reasonable short in terms of its source code.'
	doc 'It is sometimes useful as a first way to get a quick view of the contents of a Verse server.'
	doc '</p>'
	doc '<dl>'
	doc '<dt><a href="render" type="application">Simple Renderer</a> - <a href="docs/render/index.html">Documentation</a><dd>The simple renderer.'
	doc 'Press <tt>Escape</tt> or close the window to exit.'
	doc '</dl>'
	doc '</p>'
}

function dist-vml {
	( cd $ROOT/vml && make clean && make )
	FILES="vml/loader vml/saver"
	dist "$FILES"
	cp $ROOT/vml/README.html $OUT/docs/README.loadsave
	cp -r $ROOT/vml/scenes $OUT/

	doch 'VML Saver and Loader' 'vml'
	doc '<p>'
	doc 'These are tools for saving and loading the contents of a Vere server to/from the simple XML format called VML. These tools are'
	doc 'command-line only, so there is little you can do with from a web page. Open a terminal, cd to the directory where this file'
	doc 'is located, and run the tools directly instead. Please read the <a href="docs/README.loadsave">readme</a> first. A number of '
	doc 'test scenes are in the <a href="scenes/">scenes/</a> directory.'
	doc '</p>'
}

function dist-purple {
	( cd $ROOT/purple && make && cd docs/api && make clean && make)
	cp $ROOT/purple/purple $ROOT/purple/purple.h $OUT
	mkdir $OUT/plugins
	cp -r $ROOT/purple/plugins/lib*.so $OUT/plugins
	cp -r $ROOT/purple/docs/api/html $OUT/docs/purple

	doch 'Purple' 'purple'
	doc '<p>'
	doc 'This is Purple, a data-flow oriented scripting engine for Verse.'
	doc '<dl>'
	doc '<dt><a href="purple">Purple Engine</a> - <a href="docs/purple/index.html">Documentation</a></dt>'
	doc '<dd>The Purple engine itself. Starting Purple will not present you with a visible interface,'
	doc 'you need to use a dedicated user-interface client (such as Puyo, below) for that.</dd>'
	doc '</dl>'
	doc '</p>'
}

function dist-puyo {
	( cd $ROOT/puyo/docs && make clean && make )
	FILES="puyo/*.py"
	dist $FILES
	cp -r $ROOT/puyo/docs $OUT/docs/puyo

	doch 'Puyo' puyo
	doc '<p>'
	doc 'This is Puyo, a user interface client for Purple (above). To use it, you will also need to have started a'
	doc 'Purple engine.'
	doc '<dl>'
	doc '<dt><a href="puyo">Puyo</a> - <a href="docs/puyo/index.html">Documentation</a></dt>'
	doc '<dd>A user interface client for Purple. Start this once you have Purple running.</dd>'
	doc '</dl>'
}

function dist-quel {
	( cd $ROOT/quelsolaar && make )
	FILES="quelsolaar/connector quelsolaar/loqairou quelsolaar/quelsolaar quelsolaar/lpaint quelsolaar/uvedit"
	DOCS="quelsolaar/docs/connector quelsolaar/docs/loq-airou"
	dist "$FILES" "$DOCS"

	doch 'Quel Solaar Suite' 'quelsolaar'
	doc '<p>'
	doc 'This is a suite of applications that have been developed to exploit the Verse system as much as possible.'
	doc '<dl>'
	doc '<dt><a href="connector">Connector</a> - <a href="docs/connector/index.html">Documentation</a></dt>'
	doc '<dd>This is Connector, a sort of schematic data inspection tool for Verse. It is very useful.</dd>'
	doc '<dt><a href="loqairou">Loq Airou</a> - <a href="docs/loqairou/index.html">Documentation</a></dt>'
	doc '<dd>This is Loq Airou, a 3D sketching application. It can be used as a general-purpose modelling tool, too.'
	doc 'Loq Airou has built-in support for Verse''s subdivision surface primitives, and an innovative user interface.</dd>'
	doc '<dt><a href="quelsolaar">Quel Solaar</a></dt>'
	doc '<dd>Quel Solaar is a stand-alone rendering client for Verse, that simply visualizes the contents of a Verse host'
	doc 'and lets you navigate around in it. It does high-quality rendering, implementing many features of the Verse material node.</dd>'
	doc '<dt><a href="uvedit">UV Editor</a></dt>'
	doc '<dd>This a UV mapping editor, useful for artists that want to texture-map a model. It is not very stable, unfortunately.</dd>'
	doc '<dt><a href="lpaint">Layer Paint</a></dt>'
	doc '<dd>This is a layer painting tool, with which you can assign values to non-visual layers of a Verse geometry node.</dd>'
	doc '</dl>'
	doc '</p>'
}

rm -f -r $OUT && mkdir $OUT
if ! [ -d $OUT ]; then
	echo "Couldn't create output directory $OUT, aborting"
	exit
fi

mkdir $OUT/docs
if ! [ -d $OUT/docs ]; then
	echo "Couldn't create documentation directory $OUT/docs, aborting"
	exit
fi

doc-header

# Call the application distribution functions. They basically just copy stuff.
dist-verse
dist-simple
dist-vml
dist-purple
dist-puyo
dist-quel

doc-footer

# Do some cleaning.
if [ -d $OUT ]; then
	echo "Cleaning out any CVS directories ..."
	# The -prune prevents find from trying to enter the deleted directory. Nice.
	find $OUT -name 'CVS' -prune -exec rm -r \{\} \;
#	echo "Cleaning out any Makefiles ..."
#	find $OUT -name 'Makefile*' -exec rm \{} \;
	echo "Building final release archive ..."
	tar czf $OUT.tar.gz $OUT
	echo "Done, $(ls -l $OUT.tar.gz | cut -d ' ' -f5) bytes"
fi
