#!/usr/bin/env python
#
# This is "VCall", a Verse Method Call user interface client. The basic purpose is to present any
# method groups found in the object nodes on a Verse server, and then let you issue calls to the
# methods contained in them. Before issuing a call, you can assign values to the method's arguments
# by using a standard graphical user interface.
#
# Implementation notes: this of course requires PyVerse, the Python bindings for the Verse API, by
# Brecht Van Lommel. See his web page at <http://users.pandora.be/blendix/verse/> for details.
#
# Copyright (c) 2006 by PDC, KTH. Written by Emil Brink. Released under the FreeBSD license.
#

import sys

import pygtk
pygtk.require("2.0")
import gtk
import gobject

import verse

class Method:
	"""A method, which knows enough about itself to issue calls. Can also construct GTK+ UI, for setting arguments."""

	def __init__(self, group, method_id, name, args):
		self.group = group
    		self.method_id = method_id
		self.name = name
		self.args = args
		self.values = self._values_default()
		self.ui = None
		self.page = -1

	def reconfig(self, name, args):
		self.name = name
		self.ui = None
		self.page = -1
		self.args = args
		self.values = self._values_default()

	def in_node(self, node):
		return self.group.node == node

	def _matrix_default(self, size):
		"""Create a default matrix, represented as a tuple of length size * size."""
		x = []
		for i in xrange(size):
			for j in xrange(size):
				if i == j: x += [1.0]
				else: x += [0.0]
		return tuple(x)

	def _values_default(self):
		"""Create list of default values, one for each of this method's arguments. In the same order as the args."""
		v = []
		for a in self.args:
			t = a[1]
			if t >= verse.O_METHOD_PTYPE_INT8 and t <= verse.O_METHOD_PTYPE_UINT32:
				v += [0]
			elif t == verse.O_METHOD_PTYPE_REAL32:
				v += [0.0]
			elif t == verse.O_METHOD_PTYPE_REAL64:
				v += [0.0]
			elif t >= verse.O_METHOD_PTYPE_REAL32_VEC2 and t <= verse.O_METHOD_PTYPE_REAL32_VEC4:
				v += [tuple([0.0] * (t - 6))]
			elif t >= verse.O_METHOD_PTYPE_REAL64_VEC2 and t <= verse.O_METHOD_PTYPE_REAL64_VEC4:
				v += [tuple([0.0] * (t - 9))]
			elif t >= verse.O_METHOD_PTYPE_REAL32_MAT4 and t <= verse.O_METHOD_PTYPE_REAL32_MAT16:
				v += [self._matrix_default(t - 12)]
			elif t >= verse.O_METHOD_PTYPE_REAL64_MAT4 and t <= verse.O_METHOD_PTYPE_REAL64_MAT16:
				v += [self._matrix_default(t - 15)]
			elif t == verse.O_METHOD_PTYPE_STRING:
				v += [""]
			elif t == verse.O_METHOD_PTYPE_NODE:
				v += [0]
			elif t == verse.O_METHOD_PTYPE_LAYER:
				v += [0]
			else:
				print "Can't create default value for argument of type", t
		return v

	def _value_set(self, index, v):
		"""All value changes go through here, which might be useful at some point."""
		self.values[index] = v

	def evt_int_changed(self, wid, index):
		self._value_set(index, long(wid.get_value()))

	def evt_uint_changed(self, wid, index):
		self._value_set(index, long(wid.get_value()))

	def evt_float_changed(self, wid, index):
		self._value_set(index, float(wid.get_value()))

	def evt_float_vec_changed(self, wid, index, pos):
		nv = list(self.values[index])
		nv[pos] = float(wid.get_value())
		self._value_set(index, tuple(nv))

	def evt_float_mat_changed(self, wid, index, col, row, size):
		nv = list(self.values[index])
		nv[row * size + col] = float(wid.get_value())
		self._value_set(index, tuple(nv))

	def evt_string_changed(self, wid, index):
		self._value_set(index, wid.get_text())

	def _spin_real32(self, default = 0.0):
		v = gtk.SpinButton(gtk.Adjustment(default, -3.4028e38, 3.4028e38, 1, 1, 0))
		v.set_digits(7)
		v.set_alignment(1.0)
		return v

	def _spin_real64(self, default = 0.0):
		v = gtk.SpinButton(gtk.Adjustment(default, -1.7976E+308, 1.7976E+308, 1, 1, 0))
		v.set_digits(9)
		v.set_alignment(1.0)
		return v

	def _type_to_name(self, mt):
		"""Map a method parameter type into a textual name of that type."""
		return ["int8", "int16", "int32", \
			"uint8", "uint16", "uint32", \
			"real32", "real64", \
			"real32_vec2", "real32_vec3", "real32_vec4", \
			"real64_vec2", "real64_vec3", "real64_vec4", \
			"real32_mat4", "real32_mat9", "real32_mat16", \
			"real64_mat4", "real64_mat9", "real64_mat16", \
			"string", "node", "layer"][mt]

	def ui_build(self):
		"""Build user interface for the method's parameters. Simply picks a suitable (set of) widget(s) for each"""
		"""argument, based on the type."""
		vbox = gtk.VBox(False, 0)
		table = gtk.Table(len(self.args), 2)
		y = 0
		wid = None
		for a in self.args:
			t = a[1]	
			lab = gtk.Label("<small>(" + self._type_to_name(t) + ")</small> <b>" + a[0] + "</b>:")	# Fancy!
			lab.set_use_markup(True)
			lab.set_alignment(1.0, 0.5)
			table.attach(lab, 0, 1, y, y+1)
			if t >= verse.O_METHOD_PTYPE_INT8 and t <= verse.O_METHOD_PTYPE_INT32:
				wid = gtk.SpinButton(gtk.Adjustment(self.values[y], [-128, -32768, -2147483648][t], [127, 32767, 2147483647][t], 1, 1 << t, 0))
				wid.set_alignment(1.0)
				wid.connect("value_changed", self.evt_int_changed, y)
			elif t >= verse.O_METHOD_PTYPE_UINT8 and t <= verse.O_METHOD_PTYPE_UINT32:
				wid = gtk.SpinButton(gtk.Adjustment(self.values[y], 0, [255, 32767, 4294967295][t - 3], 1, 1 << (t - 3), 0))
				wid.set_alignment(1.0)
				wid.connect("value_changed", self.evt_uint_changed, y)
			elif t == verse.O_METHOD_PTYPE_REAL32:
				wid = self._spin_real32(self.values[y])
				wid.connect("value_changed", self.evt_float_changed, y)
			elif t == verse.O_METHOD_PTYPE_REAL64:
				wid = self._spin_real64(self.values[y])
				wid.connect("value_changed", self.evt_float_changed, y)
			elif t >= verse.O_METHOD_PTYPE_REAL32_VEC2 and t <= verse.O_METHOD_PTYPE_REAL32_VEC4:
				wid = gtk.VBox(False, 0)
				for i in  xrange(0, t - 6):
					v = self._spin_real32(self.values[y][i])
					v.connect("value_changed", self.evt_float_vec_changed, y, i)
					wid.pack_start(v, False, False, 0)
			elif t >= verse.O_METHOD_PTYPE_REAL64_VEC2 and t <= verse.O_METHOD_PTYPE_REAL64_VEC4:
				wid = gtk.VBox(False, 0)
				for i in  xrange(0, t - 9):
					v = self._spin_real64(self.values[y][i])
					v.connect("value_changed", self.evt_float_vec_changed, y, i)
					wid.pack_start(v, False, False, 0)
			elif t >= verse.O_METHOD_PTYPE_REAL32_MAT4 and t <= verse.O_METHOD_PTYPE_REAL32_MAT16:
				l = (t - 12)
				wid = gtk.Table(l, l)
				for i in xrange(l):
					for j in xrange(l):
						v = self._spin_real32(self.values[y][i * l + j])
						v.connect("value_changed", self.evt_float_mat_changed, y, j, i, l)
						wid.attach(v, j, j + 1, i, i + 1)
			elif t >= verse.O_METHOD_PTYPE_REAL64_MAT4 and t <= verse.O_METHOD_PTYPE_REAL64_MAT16:
				l = (t - 15)
				wid = gtk.Table(l, l)
				for i in xrange(l):
					for j in xrange(l):
						v = self._spin_real64(self.values[y][i * l + j])
						v.connect("value_changed", self.evt_float_mat_changed, y, j, i, l)
						wid.attach(v, j, j + 1, i, i + 1)
			elif t == verse.O_METHOD_PTYPE_STRING:
				wid = gtk.Entry()
				wid.set_text(self.values[y])
				wid.connect("changed", self.evt_string_changed, y)
			elif t == verse.O_METHOD_PTYPE_NODE:
				wid = gtk.SpinButton(gtk.Adjustment(self.values[y], 0, 4294967295, 1, 4, 0))
				wid.set_alignment(1.0)
				wid.connect("value_changed", self.evt_uint_changed, y)
			elif t == verse.O_METHOD_PTYPE_LAYER:
				wid = gtk.SpinButton(gtk.Adjustment(self.values[y], 0, 65535, 1, 2, 0))
				wid.set_alignment(1.0)
				wid.connect("value_changed", self.evt_uint_changed, y)
			if wid != None:
				table.attach(wid, 1, 2, y, y + 1)
				wid = None
			else:
				print "No UI widget for argument of type", t
			y += 1
		vbox.pack_start(table, False, False, 0)
		return vbox

	def ui_to_notebook(self, nbook):
		if self.ui == None:
			self.ui = self.ui_build()
			self.page = nbook.append_page(self.ui)
			self.ui.show_all()
		nbook.set_current_page(self.page)

	def send_call(self):
		verse.send_o_method_call(self.group.node.node_id, self.group.group_id, self.method_id, 0, self.values)

	def get_label(self):
		return "%s.%s.%s() [%u.%u.%u]" % (self.group.node.name, self.group.name, self.name, self.group.node.node_id, self.group.group_id, self.method_id)

class MethodGroup:
	def __init__(self, node, group_id, name):
		self.node = node
		self.group_id = group_id
		self.name = name
		self.methods = {}

	def method_create(self, method_id, name, args):
		m = Method(self, method_id, name, args)
		self.methods[method_id] = m
		return m

class Node:
	"""Simplistic base class for Verse nodes."""

	def __init__(self, node_id):
		self.node_id = node_id
		self.name = ""

	def name_set(self, name):
		self.name = name

class NObject(Node):
	"""A representation of an object. Doesn't care about rendering-oriented features, only stores method groups."""

	def __init__(self, node_id):
		Node.__init__(self, node_id)
		self.groups = {}

	def method_group_create(self, group_id, name):
		grp = MethodGroup(self, group_id, name)
		self.groups[group_id] = grp
		return grp

	def method_group_destroy(self, group_id):
		if self.groups.has_key(group_id):
			del self.groups[group_id]

	def method_create(self, group_id, method_id, name, args):
		try:	grp = self.groups[group_id]
		except:	return
		return grp.method_create(method_id, name, args)

class VCall:
	def node_lookup(self, node_id):
		try:	return self.nodes[node_id]
		except:	return None

	def store_node_get_iter(self, node_id):
		try:	iter = self.store.get_iter(0)
		except:	return None

		while iter != None and self.store.iter_is_valid(iter):
			nnid = self.store.get(iter, 0)[0]
			if nnid == node_id:
				return iter
			iter = self.store.iter_next(iter)
		return None

	def store_methodgroup_get_iter(self, node_id, group_id):
		node = self.store_node_get_iter(node_id)
		if node == None: return None
		iter = self.store.iter_children(node)
		if iter == None: return None

		while iter != None and self.store.iter_is_valid(iter):
			nnid = self.store.get(iter, 0)[0]
			if nnid == group_id:
				return iter
			iter = self.store.iter_next(iter)
		return None

	def store_method_get_iter(self, node_id, group_id, method_id):
		mg = self.store_methodgroup_get_iter(node_id, group_id)
		if mg == None: return None
		iter = self.store.iter_children(mg)
		if iter == None: return None

		while iter != None and self.store.iter_is_valid(iter):
			mid = self.store.get(iter, 0)[0]
			if mid == method_id:
				return iter
			iter = self.store.iter_next(iter)
		return None

	def store_node_update(self, node_id, name, n):
		iter = self.store_node_get_iter(node_id)
		if iter == None:
			self.store.append(None, [node_id, name, n])
			return
		self.store.set(iter, 1, name, 2, n)

	def cb_o_method_destroy(self, node_id, group_id, method_id):
		old = self.store_method_get_iter(node_id, group_id, method_id)
		if old == None: return
		m = self.store.get(old, 2)[0]
		self.store.remove(old)
		if self.method == m:
			self.select_none()

	def cb_o_method_create(self, node_id, group_id, method_id, name, args):
		n = self.node_lookup(node_id)
		if n != None:
			m = n.method_create(group_id, method_id, name, args)
			parent = self.store_methodgroup_get_iter(node_id, group_id)
			if parent != None:
				old = self.store_method_get_iter(node_id, group_id, method_id)
				if old != None:	# Rename?
					self.store.set(old, 1, name)
					m = self.store.get(old, 2)[0]
					m.reconfig(name, args)
					if self.method == m:
						self.mlabel.set_text(m.get_label())
						m.ui_to_notebook(self.nbook)
				else:
					self.store.append(parent, [method_id, name, m])

	def cb_o_method_group_destroy(self, node_id, group_id):
		n = self.node_lookup(node_id)
		if n != None:
			if self.method != None and self.method.group.group_id == group_id:
				self.select_none()
			n.method_group_destroy(group_id)
			parent = self.store_methodgroup_get_iter(node_id, group_id)
			self.store.remove(parent)

	def cb_o_method_group_create(self, node_id, group_id, name):
		verse.send_o_method_group_subscribe(node_id, group_id)
		n = self.node_lookup(node_id)
		if n != None:
			mg = n.method_group_create(group_id, name)
			parent = self.store_node_get_iter(node_id)
			if parent != None:
				old = self.store_methodgroup_get_iter(node_id, group_id)
				if old != None:	# Rename?
					self.store.set(old, 1, name)
				else:
					self.store.append(parent, [group_id, name, mg])

	def cb_node_name_set(self, node_id, name):
		n = self.node_lookup(node_id)
		if n != None:
			n.name_set(name)
			self.store_node_update(node_id, name, n)

	def cb_node_destroy(self, node_id):
		n = self.node_lookup(node_id)
		if n != None:
			parent = self.store_node_get_iter(node_id)
			del self.nodes[node_id]
			self.store.remove(parent)
			if self.method != None and self.method.in_node(n):
				self.select_none()

	def cb_node_create(self, node_id, type, owner):
		if type == verse.OBJECT:
			n = NObject(node_id)
			self.nodes[node_id] = n
			verse.send_node_subscribe(node_id)
			self.store_node_update(node_id, "(unknown)", n)

	def cb_connect_accept(self, avatar, address, host):
		self.avatar = avatar
		verse.send_node_name_set(avatar, "vcall")
		verse.send_node_index_subscribe(1 << verse.OBJECT)

	def select_none(self):
		self.nbook.set_current_page(0)
		self.mlabel.set_text("")
		self.method = None

	def evt_selection_changed(self, sel):
		s = sel.get_selected()
		iter = s[1]
		if iter == None: return
		m = self.store.get(iter, 2)[0]
		if m == None or m.__class__ != Method:	# Filter out non-leaves.
			return
		self.mlabel.set_text(m.get_label())
		m.ui_to_notebook(self.nbook)
		self.method = m

	def evt_call_clicked(self, wid):
		"""Send a call to the current method, with the current values for the arguments."""

		if self.method == None: return
		self.method.send_call()

	def evt_win_delete(self, wid, event):
		gtk.main_quit()

	def _cb_sorter(self, model, iter1, iter2):
		col = model.get_sort_column_id()[0]
		v1 = model.get(iter1, col)
		v2 = model.get(iter2, col)
		if v1 < v2: return -1
		elif v1 > v2: return 1
		return 0

	def __init__(self, ip):
		verse.callback_set(verse.SEND_CONNECT_ACCEPT,	 	self.cb_connect_accept)
		verse.callback_set(verse.SEND_NODE_CREATE,	 	self.cb_node_create)
		verse.callback_set(verse.SEND_NODE_DESTROY,	 	self.cb_node_destroy)
		verse.callback_set(verse.SEND_NODE_NAME_SET,	 	self.cb_node_name_set)
		verse.callback_set(verse.SEND_O_METHOD_GROUP_CREATE,	self.cb_o_method_group_create)
		verse.callback_set(verse.SEND_O_METHOD_GROUP_DESTROY,	self.cb_o_method_group_destroy)
		verse.callback_set(verse.SEND_O_METHOD_CREATE,	 	self.cb_o_method_create)
		verse.callback_set(verse.SEND_O_METHOD_DESTROY,	 	self.cb_o_method_destroy)

		self.nodes = {}
		self.store = gtk.TreeStore(int, str, object)	# Id, name, some object.
		self.store.set_sort_func(0, self._cb_sorter)
		self.store.set_sort_func(1, self._cb_sorter)

		self.avatar = -1
		self.server = ip
		self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.win.connect("delete_event", self.evt_win_delete)
		self.win.set_title("VCall")
		pan = gtk.HPaned()
		self.view = gtk.TreeView(self.store)
		cell = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("ID", cell, text = 0)
		tvc.set_sort_column_id(0)
		self.view.append_column(tvc)
		cell = gtk.CellRendererText()
		tvc = gtk.TreeViewColumn("Name", cell, text = 1)
		tvc.set_sort_column_id(1)
		self.view.append_column(tvc)
		sel = self.view.get_selection()
		sel.connect("changed", self.evt_selection_changed)
		self.view.set_headers_clickable(True)
		scwin = gtk.ScrolledWindow()
		scwin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		scwin.add(self.view)
		pan.add1(scwin)
		vbox = gtk.VBox(False, 0)
		self.mlabel = gtk.Label("")
		vbox.pack_start(self.mlabel, False, False, 0)
		self.nbook = gtk.Notebook()
		self.nbook.set_show_tabs(False)
		self.nbook.set_show_border(False)
		self.nbook.append_page(gtk.Label("Please select a method to edit."))
		scwin = gtk.ScrolledWindow()
		scwin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		scwin.set_shadow_type(gtk.SHADOW_NONE)
		scwin.add_with_viewport(self.nbook)
		vbox.pack_start(scwin, True, True, 0)
		btn = gtk.Button("Send Call")
		btn.connect("clicked", self.evt_call_clicked)
		vbox.pack_start(btn, False, False, 0)
		pan.add2(vbox)
		self.win.add(pan)

		self.method = None

	def evt_timeout(self):
		"""Trampoline, gets called by GTK+ periodically, and then calls Verse to update network."""

		verse.callback_update(10000)
		return 1

	def run(self):
		verse.send_connect("vcall", "<secret>", self.server, 0)
		gobject.timeout_add(35, self.evt_timeout)
		self.win.show_all()
		gtk.main()
		if self.avatar != -1:
			verse.send_connect_terminate(self.server, "All done, signing off")

def main(args = None):
	if args is None:
		args = sys.argv
	ip = "localhost"
	for a in args[1:]:
		if a.startswith("-ip="):
			ip = a[4:]
	vc = VCall(ip)
	vc.run()

if __name__ == "__main__":
	main(sys.argv)
