/*
 * vclock.c
 * 
 * A simple Verse client that creates, and keeps up-to-date, a bitmap node
 * holding a digital clock (in HH MM SS notation).
 * 
 * Copyright (C) 2006, PDC KTH. Written by Emil Brink. Licensed under the
 * FreeBSD license.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "verse.h"

#include "font-big.xpm"

typedef struct {
	unsigned short	width, height;
	unsigned char	*data;
} DigitFont;

typedef struct {
	char		pen;
	unsigned int	color;
} Pen;

typedef struct {
	VNodeID		avatar;

	DigitFont	*font;

	char		*use_node;
	int		timeout;
	int		timed_out;
	VNodeID		target;
	unsigned int	layers;
	VLayerID	layer_id[4];
	int		layers_done;

	int		border;
	int		create_all;
	int		full_update;
	int		utc_mode;
	time_t		offset;
	time_t		last_time;
	char		last_str[16];
} MainInfo;

static unsigned char palette_find(char c, const Pen *p, size_t np)
{
	size_t	i;

	for(i = 0; i < np; i++)
	{
		if(p[i].pen == c)
			return p[i].color & 0xff;
	}
	return 0;
}

void digit_font_get_tile(const DigitFont *font, int digit, int x, int y, VNBTile *tile)
{
	int	px, py, offset;

	offset = digit * (font->width / 10);
	offset += x * VN_B_TILE_SIZE + y * font->width * VN_B_TILE_SIZE;
	for(py = 0; py < VN_B_TILE_SIZE; py++)
	{
		for(px = 0; px < VN_B_TILE_SIZE; px++)
			tile->vuint8[py * VN_B_TILE_SIZE + px] = font->data[offset + py * font->width + px];
	}
}

DigitFont * digit_font_new_from_xpm(char **xpm)
{
	unsigned int	w, h, np, cs, x, y, i;
	unsigned char	*put;
	DigitFont	*fn;
	Pen		palette[32];

	if(sscanf(xpm[0], "%u %u %u %u", &w, &h, &np, &cs) != 4)
	{
		fprintf(stderr, "Couldn't parse XPM header\n");
		return NULL;
	}
	if(np > sizeof palette / sizeof *palette || cs != 1)
	{
		fprintf(stderr, "Too many colors in XPM\n");
		return NULL;
	}
	for(i = 0; i < np; i++)
	{
		if(sscanf(xpm[1 + i], "%c g #%x", &palette[i].pen, &palette[i].color) != 2)
			fprintf(stderr, "Parse error in color %u :/\n", i);
	}
	fn = malloc(sizeof *fn + w * h);
	fn->width = w;
	fn->height = h;
	fn->data = (unsigned char *) (fn + 1);
	for(y = 0, put = fn->data; y < h; y++)
	{
		for(x = 0; x < w; x++)
			*put++ = palette_find(xpm[1 + np + y][x], palette, np);
	}
	return fn;
}

static void cb_b_layer_create(void *user, VNodeID node_id, VLayerID layer_id, const char *name, VNBLayerType type)
{
	MainInfo	*min = user;
	unsigned int	i, j, k, m;

	if(strcmp(name, "color_r") == 0)
		i = 0;
	else if(strcmp(name, "color_g") == 0)
		i = 1;
	else if(strcmp(name, "color_b") == 0)
		i = 2;
	else if(strcmp(name, "alpha") == 0)
		i = 3;
	else
		return;
	min->layer_id[i] = layer_id;
	for(j = k = m = 0; j < sizeof min->layer_id / sizeof *min->layer_id; j++)
	{
		if(min->layers & (1 << j))
		{
			k++;
			if(min->layer_id[j] != (uint16) ~0u)
				m++;
		}
	}
	min->layers_done = k == m;
}

static void create_layers(MainInfo *min, VNodeID node_id)
{
	const char	*ln[] = { "color_r", "color_g", "color_b", "alpha" };
	unsigned int	i, bw, bh;

	/* Compute the image size. Clock is 6 digits, plus two quarters for spacing. */
	bw = (int) (6.5 * min->font->width) / 10;
	bw += 2 * min->border * VN_B_TILE_SIZE;
	bh = min->font->height;
	bh += 2 * min->border * VN_B_TILE_SIZE;
	verse_send_b_dimensions_set(min->target, bw, bh, 1);
	for(i = 0; i < sizeof ln / sizeof *ln; i++)
	{
		if(((i < 3 && min->create_all) && (i < sizeof ln / sizeof *ln)) || min->layers & (1 << i))
			verse_send_b_layer_create(min->target, ~0, ln[i], VN_B_LAYER_UINT8);
	}
}

static void cb_node_name_set(void *user, VNodeID node_id, const char *name)
{
	MainInfo	*min = user;

	if(!min->use_node)
		return;
	if(strcmp(name, min->use_node) == 0)
	{
		printf("Old clock bitmap (%s) found in node %u, continuing\n", name, node_id);
		min->target = node_id;
	}
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	MainInfo	*min = user;

	if(type == V_NT_BITMAP && owner)
	{
		char	buf[32];

		if(min->use_node == NULL)
			sprintf(buf, "vclock-%u-bitmap", min->avatar);
		else
			sprintf(buf, "%s", min->use_node);
		verse_send_node_name_set(node_id, buf);
		min->target = node_id;
		verse_send_node_subscribe(min->target);
		create_layers(min, node_id);
		min->use_node = NULL;
	}
	else if(min->use_node != NULL)
	{
		verse_send_node_subscribe(node_id);
	}
}

static void cb_connect_accept(void *user, VNodeID avatar, const char *server, const uint8 *host)
{
	MainInfo	*min = user;
	char		buf[32];

	min->avatar = avatar;
	printf("Connected as %u to %s\n", avatar, server);
	sprintf(buf, "vclock-%u", avatar);
	verse_send_node_name_set(avatar, buf);
	printf(" Set avatar name to '%s'\n", buf);
	verse_send_node_index_subscribe(1 << V_NT_BITMAP);
	min->last_time = time(NULL);
	if(min->use_node == NULL)
	{
		printf("Creating new node\n");
		verse_send_node_create(~0u, V_NT_BITMAP, TRUE);
	}
}

static void update(MainInfo *min)
{
	time_t		now = time(NULL);
	struct tm	*tm;

	if(now == min->last_time)
		return;
	if(min->use_node != NULL && min->target == ~0u && min->avatar != ~0u)
	{
		if(now - min->last_time > min->timeout && !min->timed_out)
		{
			printf("The requested node (%s) hasn't been found in %d seconds -- creating new\n", min->use_node, (int) (now - min->last_time));
			verse_send_node_create(~0u, V_NT_BITMAP, TRUE);
			min->timed_out = TRUE;	/* Only try once. */
			return;
		}
	}
	if(min->target == ~0u || !min->layers_done)
		return;
	min->last_time = now;
	now += min->offset;
	if(min->utc_mode)
		tm = gmtime(&now);
	else
		tm = localtime(&now);

	if(tm != NULL)
	{
		char	buf[32];
		unsigned int	i, j, x, y, dx = min->border, dy = min->border, diw = min->font->width / (10 * VN_B_TILE_SIZE);
		VNBTile	tile;

		strftime(buf, sizeof buf, "%H %M %S", tm);
		for(i = 0; buf[i] != '\0'; i++, dx += diw)
		{
			if(buf[i] == ' ')
			{
				dx -= (diw - 1);
				continue;
			}
			if(!min->full_update && buf[i] == min->last_str[i])
				continue;
			min->last_str[i] = buf[i];
			for(y = 0; y < min->font->height / VN_B_TILE_SIZE; y++)
			{
				for(x = 0; x < diw; x++)
				{
					digit_font_get_tile(min->font, buf[i] - '0', x, y, &tile);
					for(j = 0; j < sizeof min->layer_id / sizeof *min->layer_id; j++)
					{
						if(min->layers & (1 << j))
							verse_send_b_tile_set(min->target, min->layer_id[j], dx + x, dy + y, 0, VN_B_LAYER_UINT8, &tile);
					}
				}
			}
		}
	}
}

static void usage(void)
{
	printf("Usage: vclock [-ip=HOST[:PORT]] [-border=n] [-layers=rgba] [-offset=m] [-F] [-C]\n");
	printf("Where:\n");
	printf(" -ip=HOST[:PORT] Sets the host address and port number of the Verse server.\n");
	printf(" -border=n\t Surrounds the block with n tiles of black border.\n");
	printf(" -layers=rgba\t Selects which layers (red, green, blue, alpha) to create.\n");
	printf(" -utc\t\t Makes the default current time be UTC, rather than local.\n");
	printf(" -offset=n\t Sets the offset from the current time to n minutes.\n");
	printf(" -node=NAME\t Wait for bitmap node named NAME, and use that.\n");
	printf(" -wait=n\t Wait at most n seconds for the named node, then create it if not found.\n");
	printf(" -F\t\t Always send the full clock, rather than just the changing tiles.\n");
	printf(" -C\t\t Create all RGB layers, regardless of which were selected with -layers.\n");
	printf(" -h\t\t This help text.\n");

	exit(EXIT_SUCCESS);
}

int main(int argc, char *argv[])
{
	unsigned int	i;
	const char	*ip = "localhost";
	MainInfo	min;

	min.avatar = ~0u;
	min.layers = 7;
	min.target = ~0u;
	min.layers_done = 0;
	for(i = 0; i < sizeof min.layer_id / sizeof *min.layer_id; i++)
		min.layer_id[i] = (VLayerID) ~0u;
	min.utc_mode = 0;
	min.offset = 0;
	min.border = 0;
	min.create_all = 0;
	min.full_update = 0;
	min.last_time = time(NULL);
	min.use_node = NULL;
	min.timeout = 5;
	memset(min.last_str, 0, sizeof min.last_str);
	min.timed_out = 0;

	for(i = 1; argv[i] != NULL; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			ip = argv[i] + 4;
		else if(strncmp(argv[i], "-layers=", 8) == 0)
		{
			min.layers = 0;
			if(strchr(argv[i] + 8, 'r') != NULL)
				min.layers |= 1 << 0;
			if(strchr(argv[i] + 8, 'g') != NULL)
				min.layers |= 1 << 1;
			if(strchr(argv[i] + 8, 'b') != NULL)
				min.layers |= 1 << 2;
			if(strchr(argv[i] + 8, 'a') != NULL)
				min.layers |= 1 << 3;
		}
		else if(strcmp(argv[i], "-utc") == 0)
			min.utc_mode = 1;
		else if(strncmp(argv[i], "-offset=", 8) == 0)
		{
			min.offset = strtol(argv[i] + 8, NULL, 10);
			min.offset *= 60;
		}
		else if(strncmp(argv[i], "-border=", 8) == 0)
			min.border = strtoul(argv[i] + 8, NULL, 10);
		else if(strncmp(argv[i], "-node=", 6) == 0)
			min.use_node = argv[i] + 6;
		else if(strncmp(argv[i], "-wait=", 6) == 0)
			min.timeout = strtoul(argv[i] + 6, NULL, 10);
		else if(strcmp(argv[i], "-F") == 0)
			min.full_update = 1;
		else if(strcmp(argv[i], "-C") == 0)
			min.create_all = 1;
		else if(strcmp(argv[i], "-h") == 0)
			usage();
		else
			fprintf(stderr, "vclock: Ignoring unknown option '%s'\n", argv[i]);
	}
	
	min.font = digit_font_new_from_xpm(font_big_xpm);

	verse_callback_set((void *) verse_send_connect_accept,	(void *) cb_connect_accept,	&min);
	verse_callback_set((void *) verse_send_node_create,	(void *) cb_node_create,	&min);
	verse_callback_set((void *) verse_send_node_name_set,	(void *) cb_node_name_set,	&min);
	verse_callback_set((void *) verse_send_b_layer_create,	(void *) cb_b_layer_create,	&min);

	verse_send_connect("vclock", "<secret>", ip, NULL);
	
	while(1)
	{
		verse_callback_update(250000);
		update(&min);
	}

	return EXIT_SUCCESS;
}
