/*
 * Send an o_anim_run command, to see if server works.
*/

#include <stdio.h>

#include <string.h>

#include "verse.h"

typedef struct
{
	char	label[16];
} OLink;

typedef struct NObject	NObject;

struct NObject
{
	VNodeID	id;
	char	name[16];
	OLink	link[16];

	NObject	*next;
};

/* One-dimensional animation parameters. */
typedef struct
{
	int	sent;
	char	node[16];
	char	link[16];
	uint32	time_s, time_f;
	real64	pos, speed, accel;
	real64	scale, scale_speed;
} AnimRun;

typedef struct
{
	const char	*server;
	VNodeID		avatar;
	AnimRun		run[8];
	unsigned int	run_next;

	NObject		*nodes;
} MainInfo;

/* ----------------------------------------------------------------------------------------- */

static NObject * nobj_lookup(const MainInfo *min, VNodeID node_id)
{
	NObject	*iter;

	for(iter = min->nodes; iter != NULL; iter = iter->next)
	{
		if(iter->id == node_id)
			return iter;
	}
	return NULL;
}

static NObject * nobj_lookup_named(const MainInfo *min, const char *name)
{
	NObject	*iter;

	for(iter = min->nodes; iter != NULL; iter = iter->next)
	{
		if(strcmp(iter->name, name) == 0)
			return iter;
	}
	return NULL;
}

static NObject * nobj_create(MainInfo *min, VNodeID node_id)
{
	NObject	*n;

	n = malloc(sizeof *n);
	n->id = node_id;
	n->name[0] = 0;
	memset(n->link, 0, sizeof n->link);

	n->next = min->nodes;
	min->nodes = n;

	return n;
}

/* ----------------------------------------------------------------------------------------- */

static void animrun_set_defaults(AnimRun *r)
{
	r->sent = 0;
	r->node[0] = '\0';
	r->link[0] = '\0';
	r->time_s = 0;
	r->time_f = 0;
	r->pos = 0.0;
	r->speed = 1.0;
	r->accel = 0.0;
	r->scale = 1.0;
	r->scale_speed = 0.0;
}

static void animrun_trigger(MainInfo *min)
{
	int	i, j;

	for(i = 0; i < min->run_next; i++)
	{
		AnimRun	*r = min->run + i;
		NObject	*n;

		if(r->sent)
			continue;
		if((n = nobj_lookup_named(min, r->node)) != NULL)
		{
			for(j = 0; j < sizeof n->link / sizeof *n->link; j++)
			{
				if(strcmp(n->link[j].label, r->link) == 0)
				{
					printf("Node %u link %u matches %s.%s, anim here\n",
					       n->id, j, r->node, r->link);
					verse_send_o_anim_run(n->id, j,
							      r->time_s, r->time_f,
							      1,
							      &r->pos, &r->speed, &r->accel,
							      &r->scale, &r->scale_speed);
					r->sent = 1;	/* Don't send it again. */
				}
			}
		}
		else
			printf("no object '%s' for anim %d\n", r->node, i);
	}
}

/* ----------------------------------------------------------------------------------------- */

static void cb_o_anim_run(void *user, VNodeID node_id, uint16 link_id,
			  uint32 time_s, uint32 time_f, real64 pos, real64 speed, real64 accel, real64 scale, real64 scale_speed)
{
	printf("Got o_anim_run on %u.%u, t=%u.%06u, pos=%g speed=%g accel=%g, scale=%g scale_speed=%g\n",
	       node_id, link_id, time_s, time_f, pos, speed, accel, scale, scale_speed);
}

static void cb_o_link_set(void *user, VNodeID node_id, uint16 link_id, VNodeID link, const char *label, uint32 target_id)
{
	NObject	*n;

	if((n = nobj_lookup(user, node_id)) != NULL)
	{
		if(link_id < sizeof n->link / sizeof *n->link)
		{
			strcpy(n->link[link_id].label, label);
			printf("Link %u from %u to %u is labelled '%s'\n", link_id, node_id, link, label);
			animrun_trigger(user);
		}
	}
	else
		printf("Can't set link of unknown node %u\n", node_id);
}

static void cb_node_name_set(void *user, VNodeID node_id, const char *name)
{
	NObject	*n;

	if((n = nobj_lookup(user, node_id)) != NULL)
	{
		strcpy(n->name, name);
		printf("Node %u is named '%s'\n", n->id, n->name);
		animrun_trigger(user);
	}
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	if(type == V_NT_OBJECT)
	{
		printf("There's an object node called %u\n", node_id);
		nobj_create(user, node_id);
		verse_send_node_subscribe(node_id);
	}
}

static void cb_connect_accept(const void *user, VNodeID avatar, const char *address, void *host)
{
	printf("Connected as avatar %u\n", avatar);
	((MainInfo *) user)->avatar = avatar;
	verse_send_node_name_set(avatar, "animtest");
	verse_send_node_index_subscribe(1 << V_NT_OBJECT);
}

int main(int argc, char *argv[])
{

	MainInfo	min;
	int		i;

	min.server = "localhost";
	min.avatar = ~0;
	min.run_next = 0u;

	for(i = 1; argv[i] != NULL; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			min.server = argv[i] + 4;
		else if(strncmp(argv[i], "-run=", 5) == 0)
		{
			AnimRun	*r;
			int	got;

			if(min.run_next >= sizeof min.run / sizeof *min.run)
			{
				printf("Can only run %u animations, sorry\n", sizeof min.run / sizeof *min.run);
				continue;
			}
			r = min.run + min.run_next;
			animrun_set_defaults(r);
			if((got = sscanf(argv[i] + 5, "%[^,]s,%[^,]s,%u.%u,%lg,%lg,%lg,%lg,%lg",
				  r->node, r->link,
				  &r->time_s, &r->time_f,
				  &r->pos, &r->speed, &r->accel,
				  &r->scale, &r->scale_speed)) >= 4)
			{
				printf("Okay, animation %u scheduled: node=%s link=%s time=%u.%07u, "
				       "pos=%g, speed=%g, accel=%g, scale=%g, scale_speed=%g\n", min.run_next,
				       r->node, r->link, r->time_s, r->time_f, r->pos, r->speed, r->accel,
				       r->scale, r->scale_speed);
				printf(" (got %d values, the rest are defaults)\n", got);
				min.run_next++;
			}
			else
				printf("-run parse error\n");
		}
	}

	verse_callback_set(verse_send_connect_accept,	cb_connect_accept, &min);
	verse_callback_set(verse_send_node_create,	cb_node_create, &min);
	verse_callback_set(verse_send_node_name_set,	cb_node_name_set, &min);

	verse_callback_set(verse_send_o_link_set,	cb_o_link_set, &min);
	verse_callback_set(verse_send_o_anim_run,	cb_o_anim_run, &min);

	verse_send_connect("animtest", "<secret>", min.server, NULL);
	while(1)
	{
		verse_callback_update(10000);
	}
	return EXIT_SUCCESS;
}
