/*
 * Verse audio reception client.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if __APPLE_CC__
#include <OpenAL/al.h>
#include <OpenAL/alut.h>
#else
#include <AL/al.h>
#include <AL/alut.h>
#endif

#include "verse.h"

#define	BUFFER_COUNT	2		/* How many buffers we use for OpenAL queueing. */
#define	BUFFER_SIZE	32768		/* Size of each buffer. */

typedef struct ABlock	ABlock;

struct ABlock
{
	uint32	time_s, time_f;		/* Time this block needs to start playing. */
	real64	freq;			/* Sample frequency this block is at. */
	size_t	size;			/* Size of block, in bytes. */
	ABlock	*next;			/* Next block in time. */
	void	*data;			/* Actual data content of block. */
};

typedef struct
{
	const char	*server;
	const char	*node_name;
	const char	*stream_name;

	VNodeID		node_id;
	VLayerID	stream_id;

	ABlock		*free_int16;
	ABlock		*queue_int16;

	ALuint		buffer[2], source;
	uint32		next_s, next_f;		/* Desired timestamp of first block after currently playing one. */
	unsigned int	refill_count;
} MainInfo;

static void block_alloc(MainInfo *min, VNABlockType type, size_t num)
{
	void	*data;
	size_t	bs = 0, i;
	ABlock	*block;
	void	*sample;

	if(type == VN_A_BLOCK_INT16)
		bs = VN_A_BLOCK_SIZE_INT16 * sizeof (int16);
	if(bs == 0)
	{
		printf("Can't initialize blocks for type %d, not implemented\n", type);
		return;
	}
	printf("allocating %u bytes\n", num * (sizeof (ABlock) + bs));
	data = malloc(num * (sizeof (ABlock) + bs));
	if(data == NULL)
	{
		printf("Allocation failed\n");
		return;
	}
	sample = (unsigned char *) data + num * sizeof *block;
	printf("samples begin at %u bytes into allocated memory\n", (unsigned int) (sample - data));

	block = (ABlock *) data;
	min->free_int16 = block;
	for(i = 0; i < num; i++, sample += bs)
	{
		block->time_s = block->time_f = 0;
		block->size   = bs;
		block->next   = i < num - 1 ? block + 1 : NULL;
		block->data   = sample;
		block++;
	}
	printf("free initialized, first at %p, second at %p\n", min->free_int16, min->free_int16->next);
}

/* Returns 1 if timestamp of <block> is earlier than that of <other>. */
static int less_than(const ABlock *block, const ABlock *other)
{
	if(block->time_s < other->time_s)
		return 1;
	if(block->time_s > other->time_s)
		return 0;
	return block->time_f < other->time_f;
}

/* Insert <block> into queue headed at <queue>, sorting it into place based on timestamp. */
static void block_enqueue(ABlock **queue, ABlock *block)
{
/*	printf("inserting %u.%010u: ", block->time_s, block->time_f);*/
	if(*queue == NULL)
	{
		*queue = block;
		block->next = NULL;
/*		printf("queue headed\n");*/
	}
	else
	{
		ABlock	*here, *prev;

		for(here = *queue, prev = NULL; here != NULL; prev = here, here = here->next)
		{
			if(less_than(block, here))
			{
				if(prev != NULL)
					prev->next = block;
				else
					*queue = block;
				block->next = here;
/*				printf("inserted\n");*/
				prev = NULL;
				break;
			}
		}
		if(prev != NULL)
		{
			prev->next  = block;
			block->next = NULL;
/*			printf("appended\n");*/
		}
	}
/*
	{
		ABlock	*iter;
		size_t	len = 0;

		printf("queue now: ");
		for(iter = *queue; iter != NULL; iter = iter->next, len++)
			printf("%u.%010u ", iter->time_s, iter->time_f);
		printf(" (%u blocks)\n", len);
	}
*/
}

static size_t block_num(const ABlock *head)
{
	size_t	num;

	for(num = 0; head != NULL; head = head->next)
		num++;
	return num;
}

static void cb_a_stream(void *user, VNodeID node_id, VLayerID stream_id, uint32 time_s, uint32 time_f,
			VNABlockType type, real64 frequency, const VNABlock *data)
{
	MainInfo	*min = user;

/*	printf("got stream data for %u.%010u, user=%p\n", time_s, time_f, user);*/
	if(type == VN_A_BLOCK_INT16)
	{
		ABlock	*block = min->free_int16;

		if(block != NULL)
		{
			min->free_int16 = block->next;
			block->time_s = time_s;
			block->time_f = time_f;
			block->freq   = frequency;
			memcpy(block->data, data->vint16, block->size);
			block_enqueue(&min->queue_int16, block);
			printf("Audio data block enqueued, queue len=%u\n", block_num(min->queue_int16));
/*			if(block_num(min->queue_int16) > 10) This is for testing unsubscribe, never mind.
			{
				printf("stupidly un-subscribing in the midst of everything!\n");
				verse_send_a_stream_unsubscribe(node_id, stream_id);
			}
*/		}
		else
			printf("Couldn't enqueue stream data, out of free blocks\n");
	}
}

static void cb_a_stream_create(void *user, VNodeID node_id, VLayerID stream_id, const char *name)
{
	MainInfo	*min = user;

	if(strcmp(name, min->stream_name) == 0)
	{
		printf("Stream is %u.%u\n", node_id, stream_id);
		verse_send_a_stream_subscribe(node_id, stream_id);
		min->node_id   = node_id;
		min->stream_id = stream_id;
	}
}

static void cb_a_stream_destroy(void *user, VNodeID node_id, VLayerID stream_id)
{
	printf("Destroy stream %u.%u\n", node_id, stream_id);
}

static void cb_node_name_set(void *user, VNodeID node_id, const char *name)
{
	MainInfo	*min = user;

	printf("Setting name of %u to '%s'\n", node_id, name);
	if(strcmp(name, min->node_name) == 0)
	{
		printf("Found it, listening to node %u\n", node_id);
	}
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	printf("Node %u has type %d\n", node_id, type);
	if(type == V_NT_AUDIO)
		verse_send_node_subscribe(node_id);
}

static void cb_connect_accept(void *user, VNodeID avatar, const char *address, void *host)
{
	printf("Connected, sweet\n");
	verse_send_node_index_subscribe(1 << V_NT_AUDIO);
}

static void al_check(MainInfo *min, const char *what)
{
	if(alGetError() != AL_NO_ERROR) 
	{
		printf("OpenAL problem with %s--aborting\n", what);
		exit(1);
	}
}

static void al_init(MainInfo *min)
{
	ALfloat listenerPos[] = { 0.0f, 0.0f, 4.0f };
	ALfloat listenerVel[] = { 0.0f, 0.0f, 0.0f };
	ALfloat listenerOri[] = { 0.0f, 0.0f, -1.0f, 0.0f, 1.0f, 0.0f };

/*	ALfloat source0Pos[] = { 0.0f, 0.0f, -4.0f };
	ALfloat source0Vel[] = { 0.0f, 0.0f, 0.0f };
*/
	alGetError();
	alListenerfv(AL_POSITION, listenerPos);
	alListenerfv(AL_VELOCITY, listenerVel);
	alListenerfv(AL_ORIENTATION, listenerOri);
	al_check(min, "listener");

	alGenBuffers(sizeof min->buffer / sizeof *min->buffer, min->buffer);
	al_check(min, "buffers");
	alGenSources(1, &min->source);
	al_check(min, "source");
	alSource3f(min->source, AL_POSITION,        0.0, 0.0, 0.0);
	alSource3f(min->source, AL_VELOCITY,        0.0, 0.0, 0.0);
	alSource3f(min->source, AL_DIRECTION,       0.0, 0.0, 0.0);
	alSourcef(min->source,  AL_ROLLOFF_FACTOR,  0.0          );
	alSourcei(min->source,  AL_SOURCE_RELATIVE, AL_TRUE);
	al_check(min,  "source setup");
}

static int is_sequence(const ABlock *block1, const ABlock *block2, uint32 *duration)
{
	double	dur;

	if(block1 == NULL || block2 == NULL)
		return 0;
	
	dur = block1->size / (2.0 * block1->freq);
	*duration = 4294967295.0 * dur;
	return block2->time_f - block1->time_f == *duration;
}

static void stream(MainInfo *min, ALuint buffer)
{
	char	buf[8192];
	size_t	put = 0, cont;
	ALuint	rate = 0;
	ABlock	*block;
	uint32	dur;

	/* Dequeue audio blocks, filling up the buffer. Assume time is right for queue head. */
	while((block = min->queue_int16) != NULL)
	{
		if(sizeof buf - put < block->size)	/* Only play whole blocks, for simplicity's sake. */
			break;
		memcpy(buf + put, block->data, block->size);
		put += block->size;
		cont = is_sequence(block, block->next, &dur);
		if(rate == 0)
			rate = block->freq;
/*		printf("block at %u.%010u copied to out buffer\n", block->time_s, block->time_f);*/
		min->next_s = block->time_s;
		min->next_f = block->time_f + dur;
		if(min->next_f < block->time_f)
			min->next_s++;
		/* Dequeue block and add to freelist. */
		min->queue_int16 = block->next;
		block->next = min->free_int16;
		min->free_int16 = block;
		if(!cont)
			break;
	}
/*	printf("buffer filled with %u bytes for replay, rate %u\n", put, rate);
	printf(" next block scheduled for %u.%010u\n", min->next_s, min->next_f);
	printf("  there are still %u blocks in queue\n", block_num(min->queue_int16));
*/	alBufferData(buffer, AL_FORMAT_MONO16, buf, put, rate);
	al_check(min, "buffer data set");
}

int main(int argc, char *argv[])
{
	static MainInfo	min;
	int		i;

	alutInit(&argc, argv);
	al_init(&min);

	min.server	= "localhost";
	min.node_name	= "audio-test";
	min.stream_name	= "ch0";
	min.free_int16  = NULL;
	min.queue_int16 = NULL;
	min.next_s = min.next_f = 0;

	for(i = 1; argv[i] != NULL; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			min.server = argv[i] + 4;
		else if(strncmp(argv[i], "-node=", 6) == 0)
			min.node_name = argv[i] + 6;
		else if(strncmp(argv[i], "-stream=", 7) == 0)
			min.stream_name = argv[i] + 7;
		else
			printf("**Unknown option '%s'\n", argv[i]);
	}
	verse_callback_set(verse_send_connect_accept,	cb_connect_accept, &min);
	verse_callback_set(verse_send_node_create,	cb_node_create, &min);
	verse_callback_set(verse_send_node_name_set,	cb_node_name_set, &min);
	verse_callback_set(verse_send_a_stream_create,	cb_a_stream_create, &min);
	verse_callback_set(verse_send_a_stream_destroy,	cb_a_stream_destroy, &min);
	verse_callback_set(verse_send_a_stream,		cb_a_stream, &min);

	verse_send_connect("arx", "<secret>", min.server, NULL);

	block_alloc(&min, VN_A_BLOCK_INT16, 1024);	/* That's a lot of buffer space. */

	while(1)
	{
		ABlock	*block;

		verse_callback_update(1000);
		if(min.next_s != 0)
		{
			ALint	processed = 0;

			alGetSourcei(min.source, AL_BUFFERS_PROCESSED, &processed);
			al_check(&min, "reading processed buffers");
			if(min.queue_int16 == NULL)	/* Out of data? */
			{
				alSourceStop(min.source);
				printf("End of playback, waiting for more...\n");
				min.next_s = min.next_f = 0;
				verse_send_a_stream_destroy(min.node_id, min.stream_id);
				continue;
			}
			else  if(min.queue_int16->time_s != min.next_s ||
				 min.queue_int16->time_f != min.next_f)
			{
				printf("Next scheduled sound isn't now, let's wait\n");
				alSourceStop(min.source);
				min.next_s = min.next_f = 0;
				continue;
			}
			for(;processed > 0; processed--)
			{
				ALuint	buffer;

				alSourceUnqueueBuffers(min.source, 1, &buffer);
				printf("Unqueued buffer %u, refilling count=%u\n", buffer, ++min.refill_count);
				stream(&min, buffer);
				alSourceQueueBuffers(min.source, 1, &buffer);
			}
		}
		else if((block = min.queue_int16) != NULL)
		{
			uint32	 time_s, time_f;

			verse_session_get_time(&time_s, &time_f);
			if(time_s >= block->time_s && time_f >= block->time_f)
			{
				stream(&min, min.buffer[0]);
				stream(&min, min.buffer[1]);
				alSourceQueueBuffers(min.source, 2, min.buffer);
				alSourcePlay(min.source);
				al_check(&min, "playback");
				min.refill_count = 0;
			}
		}
	}
	return EXIT_SUCCESS;
}
