/*
 * Verse audio transmission client.
*/

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "verse.h"

typedef struct
{
	const char	*server;
	const char	*node_name;
	const char	*stream_name;

	VNodeID		node_id;
	VLayerID	stream_id;

	double		tone;		/* Frequency of tone we're going to generate. */
	double		duration;	/* Duration of playback, in seconds. */
} MainInfo;

static void cb_a_stream_create(void *user, VNodeID node_id, VLayerID stream_id, const char *name)
{
	MainInfo	*min = user;

	if(strcmp(name, min->stream_name) == 0)
	{
		VNABlock	block;
		uint32	bs = sizeof block.vint16 / sizeof *block.vint16, i, j, time_s, time_f, num, block_s, block_f;
		double	tone = min->tone, freq = 44.1e3, t, ang;

		min->stream_id = stream_id;

		printf("Stream is %u.%u\n", node_id, stream_id);
		verse_send_a_stream_subscribe(node_id, stream_id);

		printf("Creating stream blocks of %.1f Hz sine, sampled at %.1f kHz\n", tone, freq / 1000.0);
		printf(" That means %g s of audio per block.\n", bs / freq);
		num = ceil(min->duration / (bs / freq));
		printf(" So if we want %g s, we need %u blocks.\n", min->duration, num);
		verse_session_get_time(&time_s, NULL);
		time_s += 5;
		time_f = 0u;
		printf(" First block to start replay at T=%u.%010u.\n", time_s, time_f);
		block_s = (int) (bs / freq);
		block_f = 4294967295.0 * ((bs / freq) - block_s);
		printf(" Time increments: %u,%u\n", block_s, block_f);
		for(i = 0; i < num; i++)
		{
			for(j = 0; j < bs; j++)
			{
				t   = (i * bs + j) / freq;
				ang = 2.0 * M_PI * (t * tone);
				block.vint16[j] = pow(2, CHAR_BIT * sizeof *block.vint16 - 1) * sin(ang);
			}
			verse_send_a_stream(node_id, stream_id, time_s, time_f, VN_A_BLOCK_INT16, freq, &block);
			time_s += block_s;
			time_f += block_f;
		}
	}
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	MainInfo	*min = user;

	if(type == V_NT_AUDIO && owner == VN_OWNER_MINE)
	{
		printf("Node is %u (%x)\n", node_id, node_id);
		verse_send_node_subscribe(node_id);
		verse_send_node_name_set(node_id, min->node_name);
		verse_send_a_stream_create(node_id, ~0, min->stream_name);
	}
}

static void cb_connect_accept(void *user, VNodeID avatar, const char *address, void *host)
{
	printf("Connected, creating audio node\n");
	verse_send_node_index_subscribe(1 << V_NT_AUDIO);
	verse_send_node_create(~0, V_NT_AUDIO, 1);
}

int main(int argc, char *argv[])
{
	static MainInfo	min;
	int		i;

	min.server	= "localhost";
	min.node_name	= "audio-test";
	min.stream_name	= "ch0";
	min.node_id	= ~0;
	min.stream_id	= ~0;
	min.tone	= 1000.0;		/* Default output frequency. */
	min.duration	= 0.5;

	for(i = 1; argv[i] != NULL; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			min.server = argv[i] + 4;
		else if(strncmp(argv[i], "-node=", 6) == 0)
			min.node_name = argv[i] + 6;
		else if(strncmp(argv[i], "-stream=", 7) == 0)
			min.stream_name = argv[i] + 7;
		else if(strncmp(argv[i], "-freq=", 6) == 0)
			min.tone = strtod(argv[i] + 6, NULL);
		else if(strncmp(argv[i], "-dur=", 5) == 0)
			min.duration = strtod(argv[i] + 5, NULL);
		else
			printf("**Unknown option '%s'\n", argv[i]);
	}
	verse_callback_set(verse_send_connect_accept,	cb_connect_accept, &min);
	verse_callback_set(verse_send_node_create,	cb_node_create, &min);
	verse_callback_set(verse_send_a_stream_create,	cb_a_stream_create, &min);

	verse_send_connect("atx", "<secret>", min.server, NULL);

	while(min.node_id == ~0 && min.stream_id == (VLayerID) ~0)
	{
		verse_callback_update(10000);
	}
	printf("All done, exiting\n");

/*	verse_send_connect_terminate("localhost", "All done");*/
	while(verse_session_get_size() > 10)
		verse_callback_update(10000);

	return EXIT_SUCCESS;
}
