/*
 * A quick test of audio layer functionality. Audio layers are simply storage
 * buffers for data that is "meant" to be streamed over an audio stream at one
 * point or another. It is not possible to ask the host to play an audio layer,
 * you need to do so manually by writing it to a stream.
*/

#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "verse.h"

typedef struct
{
	const char	*server;
	int		create, destroy;
	int		verbose;
	int		delay, unsub;
	uint32		clear;
	VNodeID		avatar;
	VNodeID		node_id;
	VLayerID	buffer_id;
} MainInfo;

/* ------------------------------------------------------------------------------------------------------------------------- */

static void cb_a_block_set(void *user, VNodeID node_id, VLayerID layer_id, uint32 block_id, VNABlockType type, void *data)
{
	static const size_t	blocksize[] = {
		VN_A_BLOCK_SIZE_INT8   * sizeof (int8),
		VN_A_BLOCK_SIZE_INT16  * sizeof (int16),
		VN_A_BLOCK_SIZE_INT24  * 3 * sizeof (int8),
		VN_A_BLOCK_SIZE_INT32  * sizeof (int32),
		VN_A_BLOCK_SIZE_REAL32 * sizeof (real32),
		VN_A_BLOCK_SIZE_REAL64 * sizeof (real64) 
	};
	MainInfo	*min = user;
	size_t		i;

	printf("Got block %u data, %u bytes", block_id, blocksize[type]);
	if(min->verbose)
	{
		printf(":\n");
		for(i = 0; i < blocksize[type]; i++)
			printf("%02X ", ((unsigned char *) data)[i]);
	}
	printf("\n");
}

static void cb_a_block_clear(void *user, VNodeID node_id, VBufferID buffer_id, uint32 block_id)
{
	printf("Block %u.%u.%u was cleared\n", node_id, buffer_id, block_id);
}

static void cb_a_buffer_create(void *user, VNodeID node_id, VBufferID buffer_id, const char *name, VNABlockType type, real64 frequency)
{
	MainInfo	*min = user;
	VNABlock	block;
	int		i, j, num = 8, bs = sizeof block.vint16 / sizeof *block.vint16;
	double		t, freq = frequency, tone = 440.0, ang;

	printf("There's an audio buffer %u.%u, type %d, freq %g name '%s'\n",
	       node_id, buffer_id, type, frequency, name);
	if(min->destroy)
	{
		printf("Destroying it\n");
		verse_send_a_buffer_destroy(node_id, buffer_id);
		return;
	}
	verse_send_a_buffer_subscribe(node_id, buffer_id);
	min->buffer_id = buffer_id;
	if(min->clear != 0)
	{
		for(i = 0; i < CHAR_BIT * sizeof min->clear; i++)
		{
			if(min->clear & (1 << i))
			{
				printf("Sending clear of block %u\n", i);
				verse_send_a_block_clear(node_id, buffer_id, i);
			}
		}
		return;
	}
	if(!min->create)
		return;
	for(i = 0; i < num; i++)
	{
		for(j = 0; j < bs; j++)
		{
			t   = (i * bs + j) / freq;
			ang = 2.0 * M_PI * (t * tone);
			block.vint16[j] = pow(2, CHAR_BIT * sizeof *block.vint16 - 1) * sin(ang);
		}
		{
			printf("sending block, %u samples:\n", bs);
			for(j = 0; j < bs; j++)
				printf("%x ", block.vint16[j]);
			printf("\n");
		}
		verse_send_a_block_set(node_id, buffer_id, i, VN_A_BLOCK_INT16, &block);
	}
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	MainInfo	*min = user;

	if(type == V_NT_AUDIO)
	{
		printf("Found audio node, it's %u\n", node_id);
		min->node_id = node_id;
		verse_send_node_subscribe(node_id);
		if(owner && min->create)
		{
			printf("Creating buffer\n");
			verse_send_a_buffer_create(node_id, ~0, "pureTone", VN_A_BLOCK_INT16, 44100.0);
		}
	}
}

static void cb_connect_accept(void *user, VNodeID avatar, const char *address, void *host)
{
	MainInfo	*min = user;

	printf("\n");
	printf("Connected as %u\n", avatar);
	verse_send_node_index_subscribe(1 << V_NT_AUDIO);
	min->avatar = avatar;
	if(min->create)
		verse_send_node_create(~0, V_NT_AUDIO, 1);
}

int main(int argc, char *argv[])
{
	MainInfo	min;
	int		i;
	uint32		now;

	min.server = "localhost";
	min.avatar = ~0;
	min.node_id = ~0;
	min.clear = 0;
	min.create = 0;
	min.destroy = 0;
	min.verbose = 1;
	min.delay = 0;
	min.unsub = 0;

	for(i = 1; argv[i] != NULL; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			min.server = argv[i] + 4;
		else if(strcmp(argv[i], "-create") == 0)
			min.create = 1;
		else if(strcmp(argv[i], "-destroy") == 0)
			min.destroy = 1;
		else if(strncmp(argv[i], "-clear=", 7) == 0)
		{
			char		*ptr = argv[i] + 7, *eptr;
			unsigned long	v;

			while(1)
			{
				v = strtol(ptr, &eptr, 0);
				if(ptr == eptr)
					break;
				if(v < CHAR_BIT * sizeof min.clear)
				{
					printf("clear %lu\n", v);
					min.clear |= 1 << v;
				}
				if(*eptr == ',')
					ptr = eptr + 1;
				else
					break;
			}
		}
		else if(strcmp(argv[i], "-quiet") == 0)
			min.verbose = 0;
		else if(strcmp(argv[i], "-delay") == 0)
			min.delay = 1;
		else if(strcmp(argv[i], "-unsub") == 0)
			min.unsub = 1;
	}

	verse_callback_set(verse_send_connect_accept,	cb_connect_accept, &min);
	verse_callback_set(verse_send_node_create,	cb_node_create, &min);
	verse_callback_set(verse_send_a_buffer_create,	cb_a_buffer_create, &min);
	verse_callback_set(verse_send_a_block_set,	cb_a_block_set, &min);
	verse_callback_set(verse_send_a_block_clear,	cb_a_block_clear, &min);

	verse_send_connect("audiotest", "<secret>", min.server, NULL);

	verse_session_get_time(&now, NULL);
	while(1)
	{
		verse_callback_update(10000);
		if(min.delay || min.unsub)
		{
			uint32	now2;
			verse_session_get_time(&now2, NULL);
			if(min.delay && now2 - now > 10)
			{
				VNABlock	block;
				int		i;

				printf("Time has passed, doing delayed block_set\n");
				for(i = 0; i < sizeof block.vint16 / sizeof *block.vint16; i++)
					block.vint16[i] = rand();
				verse_send_a_block_set(min.node_id, min.buffer_id, 0, VN_A_BLOCK_INT16, &block);
				min.delay = 0;
			}
			if(min.unsub && now2 - now > 4)
			{
				printf("Time has passed, unsubcribing\n");
				verse_send_a_buffer_unsubscribe(min.node_id, min.buffer_id);
				min.unsub = 0;
			}
		}
	}
}
