/*
 * A simple Verse client to download the first bitmap it finds on a server,
 * and meausure how long it takes. Since this is not concerned with displaying
 * (or even storing) the pixels, it is very straight-forward.
 * 
 * Copyright (C) 2006 by PDC, KTH. Written by Emil Brink. Released under the
 * FreeBSD license.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>

#include "verse.h"

#define	TILE_ROUND(n)	((n + VN_B_TILE_SIZE - 1) / VN_B_TILE_SIZE)

typedef struct {
	VNodeID		*data;
	size_t		size, allocated;
} NodeIDArray;

typedef enum { DONE = -1, WAITFIRST, WAITNAMED, FLUSHING, DOWNLOAD } State;

typedef struct {
	State		state;
	int		quiet;

	const char	*ip;
	unsigned int	timeout;

	const char	*node;
	VNodeID		named_node;
	unsigned int	flush;			/* Delay spent "flushing" after name-based search. */

	VNodeID		avatar;
	VNodeID		bitmap;

	unsigned int	size_tiles;
	unsigned int	expected;
	unsigned int	expected_bytes;
	unsigned int	received;
	int		started;
	struct timeval	time0;

	NodeIDArray	history;		/* Used when subscribing to a named node. */
} MainInfo;

/* ------------------------------------------------------------------------------------------------ */

static void nia_init(NodeIDArray *array, size_t initial)
{
	if(initial > 0)
		array->data = malloc(initial * sizeof *array->data);
	else
		array->data = NULL;
	array->size = 0;
	array->allocated = 0;
}

static int nia_grow(NodeIDArray *array)
{
	if(array->size + 1 >= array->allocated)
	{
		size_t	ns = array->allocated + 128;
		VNodeID	*na = realloc(array->data, ns * sizeof *array->data);
		if(na != NULL)
		{
			array->data = na;
			array->allocated = ns;
			return 1;
		}
		fprintf(stderr, "Couldn't grow NodeID array to size %u\n", ns);
		return 0;
	}
	return 1;
}

static void nia_insert(NodeIDArray *array, VNodeID node)
{
	if(array == NULL)
		return;
	if(nia_grow(array))
		array->data[array->size++] = node;
}

static VNodeID nia_index(const NodeIDArray *array, unsigned int index)
{
	if(array != NULL)
	{
		if(index < array->size)
			return array->data[index];
	}
	return ~0u;
}

static void nia_clear(NodeIDArray *array)
{
	if(array == NULL)
		return;
	if(array->data != NULL)
		free(array->data);
	array->size = array->allocated = 0u;
}

/* ------------------------------------------------------------------------------------------------ */

static size_t type_to_size(VNBLayerType type)
{
	VNBTile	dummy;

	switch(type)
	{
	case VN_B_LAYER_UINT1:	return sizeof dummy.vuint1;
	case VN_B_LAYER_UINT8:	return sizeof dummy.vuint8;
	case VN_B_LAYER_UINT16:	return sizeof dummy.vuint16;
	case VN_B_LAYER_REAL32:	return sizeof dummy.vreal32;
	case VN_B_LAYER_REAL64:	return sizeof dummy.vreal64;
	}
	return 0;
}

static const char * type_to_string(VNBLayerType type)
{
	switch(type)
	{
	case VN_B_LAYER_UINT1:	return "UINT1";
	case VN_B_LAYER_UINT8:	return "UINT8";
	case VN_B_LAYER_UINT16:	return "UINT16";
	case VN_B_LAYER_REAL32:	return "REAL32";
	case VN_B_LAYER_REAL64:	return "REAL64";
	}
	return "<unknown>";
}

static double elapsed(const struct timeval *start)
{
	struct timeval	now;

	gettimeofday(&now, NULL);
	return now.tv_sec - start->tv_sec + 1e-6 * (now.tv_usec - start->tv_usec);
}

static void cb_b_tile_set(void *user, VNodeID node_id, VLayerID layer_id, uint16 tile_x, uint16 tile_y, uint16 z, VNBLayerType type, const VNBTile *tile)
{
	MainInfo	*min = user;

	if(node_id == min->bitmap)
	{
		min->received++;
		if(min->received >= min->expected)
		{
			double	elap = elapsed(&min->time0);

			printf("\nDone, downloaded %u bytes in %.2g seconds -> %.2f MB/s\n",
			       min->expected_bytes,
			       elap,
			       min->expected_bytes / (1024.0 * 1024.0 * elap));
			min->state = DONE;
		}
		if(min->received % (min->expected / 10) == 0)
		{
			putchar('.');
			fflush(stdout);
		}
	}
}

static void cb_b_layer_create(void *user, VNodeID node_id, VLayerID layer_id, const char *name, VNBLayerType type)
{
	MainInfo	*min = user;

	if(min->state == FLUSHING)
		return;

	if(node_id == min->bitmap)
	{
		if(strcmp(name, "color_r") == 0 ||
		   strcmp(name, "color_g") == 0 ||
		   strcmp(name, "color_b") == 0)
		{
			if(!min->quiet)
				printf("Subscribing to layer %u, '%s', type %d (%s)\n", layer_id, name, type, type_to_string(type));
			verse_send_b_layer_subscribe(node_id, layer_id, 0);
			min->expected += min->size_tiles;
			if(!min->quiet)
				printf(" Expecting %u tiles (%u bytes)\n", min->expected, min->expected * type_to_size(type));
			min->expected_bytes = min->expected * type_to_size(type);
			if(!min->started)
			{
				gettimeofday(&min->time0, NULL);
				min->started = 1;
			}
		}
	}
}

static void cb_b_dimensions_set(void *user, VNodeID node_id, uint16 width, uint16 height, uint16 depth)
{
	MainInfo	*min = user;
	size_t		tiles;

	if(min->state == FLUSHING)
		return;

	tiles = TILE_ROUND(width) * TILE_ROUND(height) * depth;
	if(!min->quiet)
		printf("Bitmap %u is %ux%ux%u, %u tiles/layer\n", node_id, width, height, depth, tiles);

	if(node_id == min->bitmap)
		min->size_tiles = tiles;
}

static void cb_node_name_set(void *user, VNodeID node_id, const char *name)
{
	MainInfo	*min = user;

	if(min->state != WAITNAMED)
		return;

	if(strcmp(name, min->node) == 0)
	{
		unsigned int	i;
		VNodeID		n;

		if(!min->quiet)
			printf("Found named node, it's %u ... flushing\n", node_id);
		for(i = 0; (n = nia_index(&min->history, i)) != ~0u; i++)
			verse_send_node_unsubscribe(n);
		nia_clear(&min->history);	/* Drop the history. */
		min->named_node = node_id;
		min->started = 0;
		min->state = FLUSHING;
		gettimeofday(&min->time0, NULL);
	}
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	MainInfo	*min = user;

	if(type == V_NT_BITMAP && min->state != FLUSHING)
	{
		if(min->bitmap == ~0u && min->state == WAITFIRST)
		{
			min->state = DOWNLOAD;
			min->bitmap = node_id;
			min->started = 0;
			verse_send_node_subscribe(node_id);
			if(!min->quiet)
				printf("Downloading node %u\n", min->bitmap);
		}
		else if(min->node != NULL && min->state == WAITNAMED)
		{
			verse_send_node_subscribe(node_id);
			nia_insert(&min->history, node_id);
		}
	}
}

static void cb_connect_accept(void *user, VNodeID avatar, const char *server, const uint8 *host)
{
	MainInfo	*min = user;

	min->avatar = avatar;
	if(!min->quiet)
		printf("Connected, subscribing to bitmap nodes ...\n");
	verse_send_node_index_subscribe(1 << V_NT_BITMAP);
}

/* Update the current state, if needed. Actually, all state changes except one are handled
 * in the various Verse callbacks, so this doesn't have too much to do.
*/
static void state_update(MainInfo *min)
{
	if(min->state == FLUSHING)
	{
		if(elapsed(&min->time0) > min->flush)	/* If a while has passed, start download. */
		{
			if(!min->quiet)
				printf("Done flushing, entering DOWNLOAD state on node %u\n", min->named_node);
			verse_send_node_subscribe(min->named_node);
			min->state = DOWNLOAD;
			min->started = 0;
			min->bitmap = min->named_node;
			min->node = NULL;		/* No longer looking for a named node. */
		}
	}
}

int main(int argc, char *argv[])
{
	MainInfo	min;
	int		i;

	min.quiet = 0;
	min.ip = "localhost";
	min.node = NULL;
	min.named_node = ~0u;
	min.flush = 5;
	min.avatar = ~0u;
	min.bitmap = ~0u;
	min.expected = 0;
	min.received = 0;
	min.timeout = 100;
	nia_init(&min.history, 0);

	for(i = 1; argv[i]; i++)
	{
		if(strncmp(argv[i], "-ip=", 4) == 0)
			min.ip = argv[i] + 4;
		else if(strncmp(argv[i], "-node=", 6) == 0)
			min.node = argv[i] + 6;
		else if(strncmp(argv[i], "-flush=", 7) == 0)
			min.flush = strtoul(argv[i] + 7, NULL, 10);
		else if(strncmp(argv[i], "-update=", 8) == 0)
			min.timeout = strtoul(argv[i] + 8, NULL, 10);
		else if(strcmp(argv[i], "-q") == 0)
			min.quiet = 1;
		else
			fprintf(stderr, "bmpbench: Unknown option '%s', ignored\n", argv[i]);
	}

	min.state = min.node == NULL ? WAITFIRST : WAITNAMED;

	verse_callback_set((void *) verse_send_connect_accept,		(void *) cb_connect_accept,	&min);
	verse_callback_set((void *) verse_send_node_create,		(void *) cb_node_create,	&min);
	verse_callback_set((void *) verse_send_node_name_set,		(void *) cb_node_name_set,	&min);
	verse_callback_set((void *) verse_send_b_dimensions_set,	(void *) cb_b_dimensions_set,	&min);
	verse_callback_set((void *) verse_send_b_layer_create,		(void *) cb_b_layer_create,	&min);
	verse_callback_set((void *) verse_send_b_tile_set,		(void *) cb_b_tile_set,		&min);

	verse_send_connect("bmpbench", "<secret>", min.ip, NULL);

	while(min.state != DONE)
	{
		verse_callback_update(min.timeout);
		state_update(&min);
	}

	verse_send_connect_terminate(min.ip, "All done, signing off");
/*	while(verse_session_get_size() > 10)
		verse_callback_update(1000);
*/
	return EXIT_SUCCESS;
}
