#!/usr/bin/python
#
# A 3DS object loader for Verse, written in Python. By Emil Brink.
#
# This requires Brecht Van Lommel's PyVerse bindings, which are available
# from his page here: <http://users.pandora.be/blendix/verse/>.
#
# Also required is the PIL imaging library (Image, below).
#
# This script will always create a simple material that does plain lit texturing.
#

import array
import math
import Image
import verse as v
import os
import sys

# -------------------------------------------------------------------------------------

GLOBAL_RESCALE = 1.0	# This can be used when models are nastily big.

# Simply wait until the session's size is less than 10 bytes, this means there is very little
# outbound traffic left to send, and we (rather safely) assume those 10 bytes hold no data.
def flush(msg = None):
	if msg != None:
		print msg
	while v.session_get_size() > 10:
		v.callback_update(10000)

def print_matrix(mtx):
	for i in range(0, 4):
		line = ""
		for j in range(0, 4):
			line += " " + str(mtx[i * 4 + j])
		print line

class Color:
	"Represents a color."

	def __init__(self):
		self.red, self.green, self.blue = 0, 0, 0
	def setRGB24(self, r, g, b):
		self.red,self.green,self.blue = r / 255.0, g / 255.0, b / 255.0

class TdsMesh:
	"Represents a collection of 3DS meshes."
	def __init__(self):
		self.version   = 0
		self.objects   = {}
		self.materials = {}

class TdsMeshObject:
	"Represents a 3DS mesh object."
	def __init__(self, name):
		self.name = name
		self.xform = [0 for i in range(16)]
		self.verts = []
		self.faces = []
		self.tex_verts = []
		self.mat_groups = []
		self.smooth = []
		self.light = ()
		self.node_object_id = ~0
		self.node_geometry_id = ~0
		self.layer_map_u = ~0
		self.layer_map_v = ~0
		self.mat_tree = None

class TdsMaterial:
	"Represents a 3DS material."

	def __init__(self):
		self.name = ""
		self.colorAmbient = None
		self.colorDiffuse = None
		self.colorSpecular = None
		self.shininess = None
		self.shin2pct = None
		self.transparency = None
		self.xpfall = None
		self.refblur = None
		self.self_ilpct = None
		self.shading = 0
		self.wire_size = 1.0
		self.tex_name = ""
		self.tex_blur = 0.0
		self.node_material_id = ~0
		self.node_bitmap_id = ~0
		self.layers   = [~0, ~0, ~0]

class BinBuffer:
	"A binary buffer, with handy accessor methods."

	def __init__(self, data):
		self.data = data
		self.p = 0

	def getEnd(self):
		return self.getPosition() >= self.getSize()

	def getSize(self):
		return len(self.data)

	def getPosition(self):
		return self.p

	def getSlice(self, pos, l):
		return self.data[pos:pos + l]

	def skip(self, amount):
		self.p += amount

	def getInt8(self):
		v = self.data[self.p]
		self.p += 1
		return v

	def getInt16(self):
		low, hi = self.data[self.p], self.data[self.p + 1]
		self.p += 2
		return 256 * hi + low

	def getInt32(self):
		low, hi = self.getInt16(), self.getInt16()
		return 65536 * hi + low

	def getFloat32(self):
		tmp = array.array('f')
		tmp.fromstring(self.data[self.p:self.p + 4])
		self.p += 4
		return tmp[0]

	def getString(self, size_max):
		out = ""
		for i in range(size_max):
			here = self.data[self.p + i]
			if here == 0:
				break
			out = out + chr(here)
		self.p += len(out) + 1
		return out

	def back(self):
		self.p -= 6

def parse_3ds_prim_color24(buf, csize):
	magic,csize = buf.getInt16(), buf.getInt32()
	if magic == 0x11 and csize == 9:
		c = Color()
		c.setRGB24(buf.getInt8(), buf.getInt8(), buf.getInt8())
		return c
	return None

def parse_3ds_prim_percentage(buf, csize):
	magic,csize = buf.getInt16(), buf.getInt32()
	if magic == 0x0030 and csize == 8:
		return buf.getInt16() / 65535.0
	elif magic == 0x0031 and csize == 10:
		return buf.getFloat32()
	return None

def parse_3ds_root(array, csize, state):
	pass

def parse_3ds_version(buf, csize, state):
	state.version = buf.getInt32()

def parse_3ds_mat_name(buf, csize, state):
	state.name = buf.getString(csize)

def parse_3ds_mat_ambient(buf, csize, state):
	state.colorAmbient = parse_3ds_prim_color24(buf, csize)

def parse_3ds_mat_diffuse(buf, csize, state):
	state.colorDiffuse = parse_3ds_prim_color24(buf, csize)

def parse_3ds_mat_specular(buf, csize, state):
	state.colorSpecular = parse_3ds_prim_color24(buf, csize)

def parse_3ds_mat_shininess(buf, csize, state):
	state.shininess = parse_3ds_prim_percentage(buf, csize)

def parse_3ds_mat_shin2pct(buf, csize, state):
	state.shin2pct = parse_3ds_prim_percentage(buf, csize)

def parse_3ds_mat_transparency(buf, csize, state):
	state.transparency = parse_3ds_prim_percentage(buf, csize)

def parse_3ds_mat_xpfall(buf, csize, state):
	state.xpfall = parse_3ds_prim_percentage(buf, csize)

def parse_3ds_mat_refblur(buf, csize, state):
	state.refblur = parse_3ds_prim_percentage(buf, csize)

def parse_3ds_mat_self_ilpct(buf, csize, state):
	state.self_ilpct = parse_3ds_prim_percentage(buf, csize)

def parse_3ds_mat_wire_size(buf, csize, state):
	state.wireSize = buf.getFloat32()

def parse_3ds_mat_shading(buf, csize, state):
	state.shading = buf.getInt16()

def parse_3ds_mat_tex_texname(buf, csize, state):
	state.tex_name = buf.getString(csize).lower()

def parse_3ds_mat_tex_tiling(buf, csize, state):
	state.tex_tiling = buf.getInt16()

def parse_3ds_mat_tex_blur(buf, csize, state):
	state.tex_blur = buf.getFloat32()

def parse_3ds_mat_texmap(buf, csize, state):
	p = parse_3ds_prim_percentage(buf, csize)
	parse_3ds(buf, csize, {
		'0xa300': parse_3ds_mat_tex_texname,
		'0xa351': parse_3ds_mat_tex_tiling,
		'0xa353': parse_3ds_mat_tex_blur,
		'0xafff': None		# Let parent handle it.
		}, state)

def parse_3ds_mat_block(buf, csize, state):
	mat = TdsMaterial()
	parse_3ds(buf, csize,
		{'0xa000': parse_3ds_mat_name,
		 '0xa010': parse_3ds_mat_ambient,
		 '0xa020': parse_3ds_mat_diffuse,
		 '0xa030': parse_3ds_mat_specular,
		 '0xa040': parse_3ds_mat_shininess,
		 '0xa041': parse_3ds_mat_shin2pct,
		 '0xa050': parse_3ds_mat_transparency,
		 '0xa052': parse_3ds_mat_xpfall,
		 '0xa053': parse_3ds_mat_refblur,
		 '0xa084': parse_3ds_mat_self_ilpct,
		 '0xa087': parse_3ds_mat_wire_size,
		 '0xa100': parse_3ds_mat_shading,
		 '0xa200': parse_3ds_mat_texmap,
		}, mat)
	state.materials[mat.name] = mat

def parse_3ds_master_scale(buf, csize, state):
	print "Master scale: " + str(buf.getFloat32())

def parse_3ds_mesh_version(buf, csize, state):
	state.version = buf.getInt32()

def parse_3ds_point_array(buf, csize, state):
	num = buf.getInt16()
	for i in range(0, num):
		x,y,z = buf.getFloat32(), buf.getFloat32(), buf.getFloat32()
		x,y,z = x,-z,y			# 3DS magic flip. Just do it.
		state.verts.append((x,y,z))
	print "Loaded " + str(len(state.verts)) + " vertices for " + state.name

def parse_3ds_face_array(buf, csize, state):
	num = buf.getInt16()
	for i in range(0, num):
		face = ( buf.getInt16(), buf.getInt16(), buf.getInt16(), buf.getInt16() )
		state.faces.append(face)
	print "Loaded", len(state.faces), "faces for", state.name

def parse_3ds_mesh_mat_group(buf, csize, state):
	mname = buf.getString(csize)
	n = buf.getInt16()
	flist = []
	for i in range(0, n):
		flist.append(buf.getInt16())
	state.mat_groups.append((mname, flist))

def parse_3ds_tex_verts(buf, csize, state):
	num = buf.getInt16()
	for i in range(0, num):
		u,v = buf.getFloat32(), buf.getFloat32()
		state.tex_verts.append(((u,v)))

def parse_3ds_smooth_group(buf, csize, state):
	for i in range(0, len(state.faces)):
		a,b = buf.getInt16(), buf.getInt16()
		state.smooth.append((a,b))

def parse_3ds_mesh_matrix(buf, csize, state):
	EPSILON = 1E-6	# Round matrix elements towards zero, for cleanliness and understanding.
	mtx = []
	for i in range(0, 12):
		mtx.append(buf.getFloat32())
		if abs(mtx[i]) < EPSILON:
			mtx[i] = 0.0
	state.xform = []
	# Assume matrix is stored as 3x3 rotation, followed by position. Convert to 4x4 form.
	pick = [ 0, 1, 2, 9,  3, 4, 5, 10,  6, 7, 8, 11 ]
	for p in pick:
		state.xform.append(mtx[p])
	state.xform[12:16] = [0.0, 0.0, 0.0, 1.0]

def parse_3ds_direct_light(buf, csize, state):
	state.light = (buf.getFloat32(), buf.getFloat32(), buf.getFloat32())

def parse_3ds_named_tri_object(buf, csize, state):
	parse_3ds(buf, csize,
		{
		'0x4110': parse_3ds_point_array,
		'0x4120': parse_3ds_face_array,
		'0x4130': parse_3ds_mesh_mat_group,
		'0x4140': parse_3ds_tex_verts,
		'0x4150': parse_3ds_smooth_group,
		'0x4160': parse_3ds_mesh_matrix,
		'0x4600': parse_3ds_direct_light,
		}, state)

def parse_3ds_named_object(buf, csize, state):
	p0 = buf.getPosition()
	name = buf.getString(csize)
	print "Parsing object " + name
	l = buf.getPosition() - p0
	csize -= l
	obj = TdsMeshObject(name)
	parse_3ds(buf, csize, { '0x4100': parse_3ds_named_tri_object }, obj)
	state.objects[name] = obj

def parse_3ds_mesh(buf, csize, state):
	mesh = TdsMesh()
	parse_3ds(buf, csize,
		{'0x3d3e': parse_3ds_mesh_version,
		 '0x100':  parse_3ds_master_scale,
		 '0x4000': parse_3ds_named_object,
		 '0xafff': parse_3ds_mat_block
		}, mesh)
	state.meshes.append(mesh)

def parse_3ds(buf, size, dict, state):
	end = buf.getPosition() + size
	while(buf.getPosition() < end):
		magic,csize = hex(buf.getInt16()), buf.getInt32()
		if dict.has_key(magic):
			if dict[magic] == None:		# Magic used to break out and let parent (if any) try it.
				buf.back()
				return
			dict[magic](buf, csize - 6, state)
		else:
			print " I cannot parse " + magic + ", skipping " + str(csize) + " bytes"
			buf.skip(csize - 6)

class TdsFile:
	"A 3ds loader class."

	def __init__(self, filename):
		try:
			f = file(filename, "rb", 32 * 1024)
			f.seek(0, 2)		# Get size of file. Idiomatic?
			size = f.tell()
			f.seek(0, 0)
			a = array.array('B')
			a.fromfile(f, size)
			f.close()
			buf = BinBuffer(a)
			self.version = 0
			self.meshes = []
			self.mat_tree = None
			# Get list of TdsMeshes
			parse_3ds(buf, size,
				{
				'0x4d4d': parse_3ds_root,
				'0x2':    parse_3ds_version,
				'0x3d3d': parse_3ds_mesh }, self)
			nv = nf = 0
			for mesh in self.meshes:
				for o in mesh.objects.itervalues():
					nv += len(o.verts)
					nf += len(o.faces)
			print "Parse complete; loaded " + str(nv) + " vertices and " + str(nf) + " faces"
		except IOError, e:
			print str(e)

# ---------------------------------------------------------------------------------------------

def test_mesh_created(f):
	for m in f.meshes:
		for o in m.objects.keys():
			if m.objects[o].node_object_id == ~0:
				return 0
	return 1

def test_geometry_created(f):
	for m in f.meshes:
		for o in m.objects.itervalues():
			if o.node_geometry_id == ~0:
				return 0
			if len(o.tex_verts) > 0:
				if o.layer_map_u == ~0:
					return 0
				if o.layer_map_v == ~0:
					return 0
	return 1

def test_materials_created(f):
	for m in f.meshes:
		for m in m.materials.itervalues():
			if m.node_material_id == ~0:
				return 0
			if m.tex_name != "" and m.node_bitmap_id == ~0:
				return 0
			break
	return 1

# ---------------------------------------------------------------------------------------------

print "Running with PyVerse r" + str(v.RELEASE_NUMBER) + "p" + str(v.RELEASE_PATCH)
for a in sys.argv[1:]:
	the_file = TdsFile(a)

	for m in the_file.meshes:
		for o in m.objects.itervalues():
			if len(o.mat_groups) > 1:
				print "Object " + o.name + " uses " + str(len(o.mat_groups)) + " materials; not supported"

	# -------------------------------------------------------------------------------------

	def cb_g_layer_create(node_id, layer_id, name, type, def_int, def_real):
		for m in the_file.meshes:
			for o in m.objects.itervalues():
				if o.node_geometry_id != node_id:
					continue;
				if name == "map_u" and o.layer_map_u == ~0:
					o.layer_map_u = layer_id
				elif name == "map_v" and o.layer_map_v == ~0:
					o.layer_map_v = layer_id
#				v.send_g_layer_subscribe(node_id, layer_id, v.FORMAT_REAL32)

	def cb_b_layer_create(node_id, layer_id, name, type):
		for mesh in the_file.meshes:
			for m in mesh.materials.itervalues():
				if m.node_bitmap_id == ~0:
					continue
				if name == "col_r" and m.layers[0] == ~0:
					m.layers[0] = layer_id
				elif name == "col_g" and m.layers[1] == ~0:
					m.layers[1] = layer_id
				elif name == "col_b" and m.layers[2] == ~0:
					m.layers[2] = layer_id
				else:
					continue
				return

	def cb_m_fragment_create(node_id, fragment_id, type, frag):
		the_file.mat_tree.upload_notify_create(fragment_id, type, frag)

	def quaternion_length(q):
		return (q[0] * q[0] + q[1] * q[1] + q[2] * q[2] + q[3] * q[3]) ** 0.5

	def quaternion_normalize(q):
		l = quaternion_length(q)
		return (q[0] / l, q[1] / l, q[2] / l, q[3] / l)

	# Convert rotational part of 4x4 matrix mtx (i.e. upper left 3x3 submatrix) to quaternion on the form (x,y,z,w).
	# Adapted from <http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/>,
	def quaternion_from_matrix(mtx):
                trace = mtx[4*0+0] + mtx[4*1+1] + mtx[4*2+2] + 1.0;
                if trace > 1e-4:
                        s = 0.5 / (trace ** 0.5)
                        x = (mtx[4*2+1] - mtx[4*1+2]) * s
                        y = (mtx[4*0+2] - mtx[4*2+0]) * s
                        z = (mtx[4*1+0] - mtx[4*0+1]) * s
                        w = 0.25 / s
                else:
                        if mtx[4*0+0] > mtx[4*1+1] and mtx[4*0+0] > mtx[4*2+2]:
                                s = 2.0 * ( 1.0 + mtx[4*0+0] - mtx[4*1+1] - mtx[4*2+2]) ** 0.5
                                x = 0.25 * s
                                y = (mtx[4*0+1] + mtx[4*1+0] ) / s
                                z = (mtx[4*0+2] + mtx[4*2+0] ) / s
                                w = (mtx[4*1+2] - mtx[4*2+1] ) / s
                        elif mtx[4*1+1] > mtx[4*2+2]:
                                s = 2.0 * ( 1.0 + mtx[4*1+1] - mtx[4*0+0] - mtx[4*2+2]) ** 0.5
                                x = (mtx[4*0+1] + mtx[4*1+0] ) / s
                                y = 0.25 * s
                                z = (mtx[4*1+2] + mtx[4*2+1] ) / s
                                w = (mtx[4*0+2] - mtx[4*2+0] ) / s
                        else:
                                s = 2.0 * ( 1.0 + mtx[4*2+2] - mtx[4*0+0] - mtx[4*1+1] ) ** 0.5
                                x = (mtx[4*0+2] + mtx[4*2+0] ) / s
                                y = (mtx[4*1+2] + mtx[4*2+1] ) / s
                                z = 0.25 * s
                                w = (mtx[4*0+1] - mtx[4*1+0] ) / s
		return (x,y,z,w)

	def cb_node_create(node_id, type, owner):
		global avatar
		if node_id == avatar:
			return
		if owner != v.OWNER_MINE:
			return
		print "node", node_id, "type", type, "created"
		if type == v.OBJECT:
			for m in the_file.meshes:
				for o in m.objects.itervalues():
					if o.node_object_id == ~0:
						o.node_object_id = node_id
						v.send_node_name_set(node_id, o.name)
						pos = (o.xform[3],o.xform[7],o.xform[11])
						rot = quaternion_normalize(quaternion_from_matrix(o.xform))
						v.send_o_transform_pos_real32(node_id, 0, 0, pos, (0,0,0),   (0,0,0),   (0,0,0),   0)
						v.send_o_transform_rot_real32(node_id, 0, 0, rot, (0,0,0,0), (0,0,0,0), (0,0,0,0), 0)
						return
		elif type == v.GEOMETRY:
			for m in the_file.meshes:
				for o in m.objects.itervalues():
					if o.node_geometry_id == ~0:
						o.node_geometry_id = node_id
						v.send_node_name_set(node_id, o.name + "-geo")
						print "setting geometry link from", o.node_object_id, "to", node_id
						v.send_o_link_set(o.node_object_id, ~0, node_id, "geometry", 0)
						v.send_node_subscribe(node_id)
						if len(o.tex_verts) > 0:
							v.send_g_layer_create(node_id, ~0, "map_u", v.G_LAYER_POLYGON_CORNER_REAL, 0, 0)
							v.send_g_layer_create(node_id, ~0, "map_v", v.G_LAYER_POLYGON_CORNER_REAL, 0, 0)
						return
		elif type == v.MATERIAL:
			for m in the_file.meshes:
				for m in m.materials.itervalues():
					if m.node_material_id == ~0:
						m.node_material_id = node_id
						v.send_node_subscribe(node_id)
						return
		elif type == v.BITMAP:
			for m in the_file.meshes:
				for m in m.materials.itervalues():
					if m.node_bitmap_id == ~0:
						m.node_bitmap_id = node_id
						v.send_node_name_set(node_id, m.tex_name)
						v.send_node_subscribe(node_id)
						for ch in "rgb":
							v.send_b_layer_create(node_id, ~0, "col_" + ch, v.B_LAYER_UINT8)
						return

	def cb_connect_accept(my_avatar, address, host_id):
		global avatar
		avatar = my_avatar
		print "Connected to host " + str(address) + " as avatar " + str(avatar)
		v.send_node_index_subscribe((1 << v.OBJECT) | (1 << v.GEOMETRY) | (1 << v.MATERIAL) | (1 << v.BITMAP))
		for m in the_file.meshes:
			for o in m.objects.keys():
				v.send_node_create(~0, v.OBJECT, v.OWNER_MINE)

	# Return (type,fragment) pair for material fragment of type <type> (a string). Sets fields to <fields>.
	def make_frag(type, fields):
		tname = "M_FT_" + type.upper()
		cname = "MF" + type[0].upper() + type[1:]
		code = "type = v." + tname + "\n" + "f = v." + cname + "()"
		exec code
		for (k,val) in fields.iteritems():
			f.__setattr__(k, val)
		return (type,f)

	class MatTree:
		"A Verse material node tree helper class."

		class MatNode:
			def __init__(self, list, parent = None):
				if list == []:
					return None
				self.type   = list[0].split("(")[0]
				self.vtype  = -1
				self.name   = list[0].split("(")[1].rstrip(")")
				self.parent = parent
				self.children = []
				self.frag_id = ~0
				if len(list) > 1:
					for ch in list[1:]:
						self.children.append(MatTree.MatNode(ch, self))

			def depth(self):
				if self == None or self.parent == None: 
					return 0
				return 1 + self.parent.depth()

			def do_to_list(self, flat):
				if self != None:
					d = self.depth()
					if len(flat) < d + 1:	# Massive overkill. It's a tree, after all.
						for i in range(len(flat), d + 1):
							flat.append([])
					flat[d].append(self)
					for c in self.children:
						c.do_to_list(flat)
				
			def to_list(self):
				flat = []
				self.do_to_list(flat)
				return flat

			def find_by_name(self, name):
				if self == None:
					return None
				if self.name == name:
					return self
				for c in self.children:
					r = c.find_by_name(name)
					if r != None:
						return r
				return None

		def __init__(self, list):
			self.tree = MatTree.MatNode(list)
			self.flat = self.tree.to_list()
			self.level = 0

		def upload_level(self):
			self.to_go = len(self.flat[self.level])
			for f in self.flat[self.level]:
				settings = self.setter(self, self.material, f.name)
				(type, frag) = make_frag(*settings)
				f.vtype = type
				v.send_m_fragment_create(self.material.node_material_id, ~0, type, frag)

		def upload(self, material, setter):
			self.material = material
			self.setter = setter
			self.level = len(self.flat) - 1
			self.upload_level()

		def upload_notify_create(self, fragment_id, type, frag):
			for f in self.flat[self.level]:
				if f.frag_id == ~0 and f.vtype == type:
					f.frag_id = fragment_id
			left = 0
			for f in self.flat[self.level]:
				left += f.frag_id == ~0
			if left == 0:
				if self.level > 0:
					self.level -= 1
					self.upload_level()

		def find_id(self, name):
			n = self.tree.find_by_name(name)
			if n != None:
				return n.frag_id
			return ~0

	v.callback_set(v.SEND_CONNECT_ACCEPT,			cb_connect_accept)
	v.callback_set(v.SEND_NODE_CREATE,			cb_node_create)
	v.callback_set(v.SEND_G_LAYER_CREATE,			cb_g_layer_create)
	v.callback_set(v.SEND_B_LAYER_CREATE,			cb_b_layer_create)
	v.callback_set(v.SEND_M_FRAGMENT_CREATE,		cb_m_fragment_create)

	server_address = "localhost"
	v.send_connect("load3ds", "<secret>", server_address, 0)

	# Wait for all object nodes to be created. Need replies, can't just look at session size.
	while not test_mesh_created(the_file):
		v.callback_update(10000)

	# Fine, now go ahead and create one geometry node each, and the UV layers.
    	for m in the_file.meshes:
		for o in m.objects.itervalues():
			v.send_node_create(~0, v.GEOMETRY, v.OWNER_MINE)
	# And again wait for replies.
	print "Creating object and geometry nodes..."
	while not test_geometry_created(the_file):
		v.callback_update(10000)

	# Then material nodes...
	for m in the_file.meshes:
		for m in m.materials.itervalues():
			v.send_node_create(~0, v.MATERIAL, v.OWNER_MINE)
			if m.tex_name != "":
				v.send_node_create(~0, v.BITMAP, v.OWNER_MINE)
	print "Creating material and bitmap nodes..."
	while not test_materials_created(the_file):
		v.callback_update(10000)

	# Now we have all we need to actually upload data. So let's do that.
	for m in the_file.meshes:
		for o in m.objects.itervalues():
			print "Uploading " + str(len(o.verts)) + " vertices for " + str(o.node_geometry_id) + "..."
			for i in range(0, len(o.verts)):
				xyz = o.verts[i]
				v.send_g_vertex_set_xyz_real32(o.node_geometry_id, 0, i, GLOBAL_RESCALE * xyz[0], GLOBAL_RESCALE * xyz[1], GLOBAL_RESCALE * xyz[2])
#			flush()			# FIXME: I'm not sure I like this one.
	# Polygon references are fine to leave dangling, so... let them.
	for m in the_file.meshes:
		for o in m.objects.itervalues():
			print "Uploading", len(o.faces), "polygons for", o.node_geometry_id, "..."
			for i in range(0, len(o.faces)):
				f = o.faces[i]
				v.send_g_polygon_set_corner_uint32(o.node_geometry_id, 1, i, f[2], f[1], f[0], ~0)
#			flush()			# FIXME: I'm not sure I like this one.
	flush("Waiting for geometry sets...")

	# Now, we know that the polygon layers have grown, so we can efficiently set the UVs.
	for m in the_file.meshes:
		for o in m.objects.itervalues():
			if len(o.tex_verts) == 0:		# Don't try setting UVs if there are none.
				continue;
			for i in range(0, len(o.faces)):
				ref = o.faces[i]		# Get references.
				try:
					u = (o.tex_verts[ref[0]][0], o.tex_verts[ref[1]][0], o.tex_verts[ref[2]][0], o.tex_verts[ref[3]][0])
					v.send_g_polygon_set_corner_real32(o.node_geometry_id, o.layer_map_u, i, u[0], u[1], u[2], u[3])
					w = (o.tex_verts[ref[0]][1], o.tex_verts[ref[1]][1], o.tex_verts[ref[2]][1], o.tex_verts[ref[3]][1])
					v.send_g_polygon_set_corner_real32(o.node_geometry_id, o.layer_map_v, i, w[0], w[1], w[2], w[3])
				except:	print "Bad texture reference found"
	# Wait for basics to arrive back. Test is simplistic, but hey.
	flush( "Waiting for UV data...")

	# Upload bitmaps. Lots of work.
	for mesh in the_file.meshes:
		for m in mesh.materials.itervalues():  
			if m.tex_name == "":
				continue
		    	picname = os.path.join(os.path.split(a)[0], m.tex_name)
			try:
				image = Image.open(picname)
				bbox = image.getbbox()
				if bbox[0] != 0 or bbox[0] != 0:
					print "Not uploading image " + picname + ", it has weird bbox " + bbox
					continue
				def num_tiles(x):
					return (x + v.B_TILE_SIZE - 1) / v.B_TILE_SIZE
				def get_tile(img, modulo, x, y, channel):
					data = []
					raw = img.getdata()
					for i in range(0, v.B_TILE_SIZE):
						for j in range(0, v.B_TILE_SIZE):
							py = y * v.B_TILE_SIZE + i
							px = x * v.B_TILE_SIZE + j
							data.append(raw[py * modulo + px][channel])
					return data
				def send_tile(m, x, y, img, modulo, channel):
					tile = get_tile(img, modulo, x, y, channel)
					v.send_b_tile_set(m.node_bitmap_id, m.layers[channel], x, y, 0, v.B_LAYER_UINT8, tile)
				def send_tiles(img, modulo, channel, m):
					for y in range(0, num_tiles(bbox[3])):
						for x in range(0, num_tiles(bbox[2])):
							send_tile(m, x, y, image, bbox[2], channel)
					flush()
				v.send_b_dimensions_set(m.node_bitmap_id, bbox[2], bbox[3], 1)
				print "Sending bitmap " + picname + ":"
				flush()
				send_tiles(image, bbox[2], 0, m)
				send_tiles(image, bbox[2], 1, m)
				send_tiles(image, bbox[2], 2, m)
			except IOError, e:
				print "Image loading of", picname, "failed"
	flush( "Waiting for bitmap data...")

	# This contains the actual knowledge about how the material should look. A bit messy, perhaps.
	# It always emits the same, hardcoded, material; a simple lit texture map.
	def setter(tree, material, name):
		if name == "uv":
			return ("geometry", { "layer_r": "map_u", "layer_g": "map_v", "layer_b": "" })
		elif name == "mat":
			return ("texture", { "bitmap": material.node_bitmap_id,
						"layer_r": "col_r", "layer_g": "col_g", "layer_b": "col_b"})
		elif name == "sun":
			return ("light", { "type": v.M_LIGHT_DIRECT_AND_AMBIENT })
		elif name == "mixtex":
			return ("blender", { "type": v.M_BLEND_MULTIPLY,
					    "data_a": tree.find_id("sun"),
					    "data_b": tree.find_id("mat"),
					    "control": ~0})
		elif name == "color":
			return ("output", { "type": "color", "front": tree.find_id("mixtex"), "back": ~0})

	# Set object links to, and create internal structure of, material nodes.
	for mesh in the_file.meshes:
		for o in mesh.objects.itervalues():
			if len(o.mat_groups) == 0:
				continue;
			if o.mat_groups[0][0] in mesh.materials:
				m = mesh.materials[o.mat_groups[0][0]]
				v.send_o_link_set(o.node_object_id, ~0, m.node_material_id, "material", 0)
				the_file.mat_tree = MatTree(["output(color)", ["blender(mixtex)", ["light(sun)"],
										["texture(mat)", ["geometry(uv)"]]]])
				the_file.mat_tree.upload(m, setter)
			else:
				print "Unknown material " + str(o.mat_groups[0][0])
	flush("Waiting for material linking...")
	v.send_connect_terminate(server_address, "All done, bye bye")
	flush("All done, disconnecting")
