/*
 * Load a PNG image, straight into a Verse host. Uses the libpng library for PNG decoding.
 * 
 * This program happily assumes that loading an image is so quick it doesn't kill the
 * connection when done "inline".
 * 
 * Written by Emil Brink.
 * 
 * 
 * Copyright (c) 2005, PDC, KTH. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are permitted
 * provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer
 *    in the documentation and/or other materials provided with the distribution.
 *  * Neither the name of PDC, KTH nor the names of its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <png.h>

#include "verse.h"

typedef enum {
	EXIT, LOAD_PNG, CREATE_NODE, CREATE_NODE_PEND, CREATE_LAYERS,
	CREATE_LAYERS_PEND, UPLOAD_TILES, NEXT
} State;

static const char	*the_layers[] = { "color_r", "color_g", "color_b", "alpha" };	/* Maps layer name to index. */

static struct {
	char		server[64];
	char		**argv;				/* From command-line, stripped of options. */
	char		filename[1024], nodename[512];	/* Split file[] members into here. */
	int		file_count, file_now;

	VNodeID		node;
	VLayerID	layer[sizeof the_layers / sizeof *the_layers];	/* Order here is RGBA, always. */
	VNBLayerType	layertype;					/* 8 or 16 bit, uint. */

	png_uint_32	width, height;	/* PNG dimensions. */
	int		bits;		/* Bits per channel. */ 
	png_uint_32	bpp;		/* Bytes per pixel. */
	int		ctype;		/* Color type from PNG. */
	uint8		**row;		/* Allocated all as one. */

	int		quiet;
	int		benchmark;
	struct timeval	start;

	uint32		image_count;
	uint32		tile_count;

	State		state;	
} min;	/* For "maininfo". */

/* ------------------------------------------------------------------------------------------------- */

static void cb_b_layer_create(void *user, VNodeID node_id, VLayerID layer_id, const char *name, VNBLayerType type)
{
	int	i, found = 0, num;

	for(i = 0; i < sizeof the_layers / sizeof *the_layers; i++)
	{
		if(strcmp(the_layers[i], name) == 0)
			min.layer[i] = layer_id;
		if(min.layer[i] != (VLayerID) ~0u)
			found++;
	}
	num = 3 + (min.ctype == PNG_COLOR_TYPE_RGB_ALPHA);
	if(found == num)		/* Are we done? */
		min.state = UPLOAD_TILES;
}

static void cb_node_create(void *user, VNodeID node_id, VNodeType type, VNodeOwner owner)
{
	if(type == V_NT_BITMAP && owner == VN_OWNER_MINE)
	{
		if(min.state == CREATE_NODE_PEND)
		{
			min.node = node_id;
			min.state = CREATE_LAYERS;
			verse_send_node_subscribe(node_id);
			verse_send_b_dimensions_set(node_id, min.width, min.height, 1);
			verse_send_node_name_set(node_id, min.nodename);
			printf("Created node %u, set name to \"%s\"\n", node_id, min.nodename);
		}
		else
			fprintf(stderr, "pngload: Got out-of-sequence node create? Odd ...\n");
	}
}

static void cb_connect_accept(void *data, VNodeID avatar, const char *address, void *host)
{
	gettimeofday(&min.start, NULL);
}

static void forget(void)
{
	int	i;

	if(min.row != NULL)
	{
		free(min.row);
		min.row = NULL;
	}
	min.node = ~0;
	for(i = 0; i < sizeof min.layer / sizeof *min.layer; i++)
		min.layer[i] = (VLayerID) ~0;
}

static void split_name(const char *name)
{
	char	*p;

	snprintf(min.filename, sizeof min.filename, "%s", name);
	if((p = strrchr(min.filename, '@')) != NULL)
	{
		*p = '\0';	/* Terminate filename. */
		snprintf(min.nodename, sizeof min.nodename, "%s", p + 1);
	}
	else
	{
		if((p = strrchr(min.filename, '\\')) != NULL ||
		   (p = strrchr(min.filename, '/')) != NULL)
			snprintf(min.nodename, sizeof min.nodename, p + 1);
		else
			snprintf(min.nodename, sizeof min.nodename, min.filename);
		if((p = strstr(min.nodename, ".png")) != NULL)
			*p = '\0';
	}
}

/* Load the next image, and kick-start upload. */
static int load(void)
{
	FILE 		*fp;
	unsigned char	header[8];
	int		got;
	png_structp	png_ptr;
	png_infop	info_ptr;
	png_uint_32	y;
	uint8		*base;

	split_name(min.argv[min.file_now]);
	fp = fopen(min.filename, "rb");
	if(!fp)
		return 0;
	got = fread(header, 1, sizeof header, fp);
	if(png_sig_cmp(header, 0, got) != 0)
	{
		fprintf(stderr, "PNG header bad\n");
		return 1;
	}

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if(png_ptr == NULL)
	        return 0;
	info_ptr = png_create_info_struct(png_ptr);
	if(info_ptr == NULL)
	{
        	png_destroy_read_struct(&png_ptr, NULL, NULL);
	        return 0;
	}
	if(setjmp(png_jmpbuf(png_ptr)))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
		fclose(fp);
		fprintf(stderr, "pngload: PNG read error/exception caught\n");
		return 0;
	}
	
	png_init_io(png_ptr, fp);
/*	png_set_strip_16(png_ptr)*/
	png_set_sig_bytes(png_ptr, sizeof header);
	png_read_info(png_ptr, info_ptr);

	png_get_IHDR(png_ptr, info_ptr, &min.width, &min.height, &min.bits, &min.ctype, NULL, NULL, NULL);

	if(min.ctype != PNG_COLOR_TYPE_RGB && min.ctype != PNG_COLOR_TYPE_RGB_ALPHA)
	{
		fprintf(stderr, "pngload: %s has unsupported PNG color type %u--need RGB or RGB_ALPHA\n", min.argv[min.file_now], min.ctype);
		return 0;
	}
	min.row = malloc(min.height * (sizeof *min.row + info_ptr->rowbytes));
	min.bpp = info_ptr->rowbytes / min.width;
	for(y = 0, base = (uint8 *) (min.row + min.height); y < min.height; y++, base += info_ptr->rowbytes)
		min.row[y] = base;
	png_read_image(png_ptr, min.row);

	png_read_end(png_ptr, NULL);
	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

	min.layertype = min.bits == 8 ? VN_B_LAYER_UINT8 : VN_B_LAYER_UINT16;

	if(!min.quiet)
		printf("Loaded \"%s\", %ux%u pixels, %s, %u bits/pixel\n", min.argv[min.file_now], (unsigned int) min.width, (unsigned int) min.height,
		       min.ctype == PNG_COLOR_TYPE_RGB ? "RGB" : "RGBA", min.bits);

	return 1;
}

static void tile_get(VNBTile *tile, unsigned int tx, unsigned int ty, unsigned int c, unsigned int wt, unsigned int ht)
{
	int	tw, th, x, y;

	tw = (tx == wt - 1) && (min.width  % VN_B_TILE_SIZE) != 0 ? min.width  % VN_B_TILE_SIZE : VN_B_TILE_SIZE;
	th = (ty == ht - 1) && (min.height % VN_B_TILE_SIZE) != 0 ? min.height % VN_B_TILE_SIZE : VN_B_TILE_SIZE;

	if(min.layertype == VN_B_LAYER_UINT8)
	{
		const uint8	*get;

		memset(tile->vuint8, 0, sizeof tile->vuint8);	/* Loops below won't touch all pixels at border. */
		for(y = 0; y < th; y++)
		{
			get = min.row[ty * VN_B_TILE_SIZE + y] + tx * VN_B_TILE_SIZE * min.bpp + c;
			for(x = 0; x < tw; x++, get += min.bpp)
				tile->vuint8[y * VN_B_TILE_SIZE + x] = *get;
		}
	}
	else if(min.layertype == VN_B_LAYER_UINT16)
	{
		const uint16	*get;

		memset(tile->vuint16, 0, sizeof tile->vuint16);
		for(y = 0; y < th; y++)
		{
			get = (const uint16 *) (min.row[ty * VN_B_TILE_SIZE + y] + tx * VN_B_TILE_SIZE * min.bpp + 2 * c);
			for(x = 0; x < tw; x++, get += min.bpp / 2)
				tile->vuint16[y * VN_B_TILE_SIZE + x] = *get;
		}
	}
}

static void upload(void)
{
	int	ch, numch,  x, y, wt, ht;
	VNBTile	tile;

	min.image_count++;

	wt = (min.width  + VN_B_TILE_SIZE - 1) / VN_B_TILE_SIZE;
	ht = (min.height + VN_B_TILE_SIZE - 1) / VN_B_TILE_SIZE;

	numch = 3 + (min.ctype == PNG_COLOR_TYPE_RGB_ALPHA);
	for(ch = 0; ch < numch; ch++)	/* Loop over channels/layers. */
	{
		for(y = 0; y < ht; y++)
		{
			for(x = 0; x < wt; x++)
			{
				tile_get(&tile, x, y, ch, wt, ht);
				verse_send_b_tile_set(min.node, min.layer[ch], x, y, 0, min.layertype, &tile);
				min.tile_count++;
			}
		}
	}
	min.state = NEXT;
}

/* Main, state-driven, "brain" of this program. */
static void think(void)
{
	switch(min.state)
	{
	case EXIT:
		break;
	case LOAD_PNG:
		forget();
		if(load())
			min.state = CREATE_NODE;
		else
			min.state = EXIT;
		break;
	case CREATE_NODE:
		verse_send_node_create(~0, V_NT_BITMAP, VN_OWNER_MINE);
		min.state = CREATE_NODE_PEND;
		break;
	case CREATE_NODE_PEND:
		/* Nothing to do, state-switching is done by callback. */
		break;
	case CREATE_LAYERS:
		{
			int		i, num;

			num = 3 + (min.ctype == PNG_COLOR_TYPE_RGB_ALPHA);
			for(i = 0; i < num; i++)
				verse_send_b_layer_create(min.node, ~0, the_layers[i], min.layertype);
			min.state = CREATE_LAYERS_PEND;
		}
		break;
	case CREATE_LAYERS_PEND:
		/* Nothing to do, state-switching is done by callback. */
		break;
	case UPLOAD_TILES:
		upload();
		break;
	case NEXT:
		min.file_now++;
		if(min.file_now >= min.file_count)
			min.state = EXIT;
		else
			min.state = LOAD_PNG;
		break;
	}
}

static void arg_remove(int *argc, char *argv[], int index)
{
	for(; index < *argc; index++)
		argv[index] = argv[index + 1];
	(*argc)--;
}

int main(int argc, char **argv)
{
	int	i;

	strcpy(min.server, "localhost");
	min.file_count = 0;
	min.quiet = 0;
	min.benchmark = 0;

	arg_remove(&argc, argv, 0);	/* Remove initial program name. */
	for(i = 0; i < argc; i++)
	{
		if(argv[i][0] == '-')
		{
			if(strcmp(argv[i], "-b") == 0)
				min.benchmark = 1;
			else if(strstr(argv[i], "help") != NULL)
			{
				printf("Usage: pngload [-ip=HOST[:PORT]] [-q] filename[@nodename] ...\n");
				printf("Default is localhost, and the Verse standard port.\n");
				printf("Use @name suffix on filename to select name of bitmap node,\n");
				printf("if omitted filename without path and suffix is used.\n");
				return EXIT_SUCCESS;
			}
			else if(strncmp(argv[i], "-ip=", 4) == 0)
			{
				strcpy(min.server, argv[i] + 4);
			}
			else if(strcmp(argv[i], "-q") == 0)
				min.quiet = 1;
			else
				fprintf(stderr, "pngload: Don't know how to handle option '%s' -- ignored\n", argv[i]);
			arg_remove(&argc, argv, i);
			i--;
		}
	}

	verse_callback_set((void *) verse_send_connect_accept,	cb_connect_accept,	NULL);
	verse_callback_set((void *) verse_send_node_create,	cb_node_create,		NULL);
	verse_callback_set((void *) verse_send_b_layer_create,	cb_b_layer_create,	NULL);

	verse_send_connect("pngload", "<secret>", min.server, NULL);

	min.argv = argv;
	min.file_count = argc;
	min.state = LOAD_PNG;
	min.file_now = 0;
	min.row = NULL;
	min.image_count = 0u;
	min.tile_count = 0u;
	forget();

	verse_send_node_index_subscribe(1 << V_NT_BITMAP);
	
	while(min.state != EXIT)
	{
		verse_callback_update(10000);
		think();
	}

	while(verse_session_get_size() > 10)
		verse_callback_update(10000);
	verse_send_connect_terminate(min.server, "All done, signing off");

	if(!min.quiet)
		printf("Done, uploaded %u images, %u tiles in total\n", min.image_count, min.tile_count);
	if(min.benchmark)
	{
		struct timeval	now;
		double		t;

		gettimeofday(&now, NULL);
		t = (now.tv_sec - min.start.tv_sec) + 1e-6 * (now.tv_usec - min.start.tv_usec);
		printf("Time taken was %g s, or %g tiles/s (%g Bps, assuming 8 bpp tiles).\n",
		       t,
		       min.tile_count / t,
		       (VN_B_TILE_SIZE * VN_B_TILE_SIZE * min.tile_count) / t);
	}
	return EXIT_SUCCESS;
}
