#!/usr/bin/python
#
# A bitmap printing Verse client. Rough and ready, but should be a good start to actually
# export bitmaps from a host to e.g. files on disk, if desired.
#
# Note: like any other Python program using Verse, this requires Brecht Van Lommel's
# PyVerse bindings. Get them from <http://users.pandora.be/blendix/verse/>.
#

import array
import verse as v
import string

# -------------------------------------------------------------------------------------

class BitmapLayer:
	def __init__(self, node, name, type):
		self.node = node
		self.name = name
		self.type = type
		self.data = None
		self.row_size = 0
		self.plane_size = 0

	def resize(self):
		self.row_size = self.node.width
		if self.type == v.B_LAYER_UINT1:
			self.data = array.array('B')
			self.row_size = int((self.node.width + 7) // 8)
		elif self.type == v.B_LAYER_UINT8:
			self.data = array.array('B')
		elif self.type == v.B_LAYER_REAL32:
			self.data = array.array('f')
		elif self.type == v.B_LAYER_REAL64:
			self.data = array.array('d')
		self.plane_size = self.row_size * self.node.height
		size = self.plane_size * self.node.depth
		self.data.fromstring(chr(255) * self.data.itemsize * size)

	def get_pixel(self, x, y, z):
		pz = z * self.plane_size
		py = y * self.row_size
		if self.type == v.B_LAYER_UINT1:
			d = self.data[pz + py + x / 8]
			if d & (1 << 7 - (x % 8)):
				return "*"
			return " "
		elif self.type == v.B_LAYER_UINT8:
			return string.zfill(hex(self.data[pz + py + x])[2:], 2)
		return " " + str(self.data[pz + py + x])

	def __str__(self):
		line = ""
		for y in range(0, self.node.height):
			for x in range(0, self.node.width):
				line += self.get_pixel(x, y, 0)
			line += "\n"
		return line

	def tile_set(self, x, y, z, tile):
		pz = z * self.plane_size
		tw = min(v.B_TILE_SIZE, self.node.width  - x * v.B_TILE_SIZE)
		th = min(v.B_TILE_SIZE, self.node.height - y * v.B_TILE_SIZE)
#		print "tx="+str(x)+", ty="+str(y)+", tw="+str(tw) + ", th="+str(th)
#		print "Tile " + str((x,y,z)) + " in " + self.name + ":"
#		for i in range(0, v.B_TILE_SIZE):
#			line = ""
#			for j in range(0, v.B_TILE_SIZE):
#				line += " " + str(tile[i * v.B_TILE_SIZE + j])
#			print line
		if self.type == v.B_LAYER_UINT1:

			def get_pixel(tile, x, y):
				return tile[y * v.B_TILE_SIZE + x] > 0

			for i in range(0, th):
				py = (y * v.B_TILE_SIZE + i) * self.row_size
				byte = 0
				for j in range(0, tw):
					if(j > 0 and (j % 8) == 0):
						self.data[pz + py + x + j / 8] = byte
						byte = 0
					byte <<= 1;
					byte |= get_pixel(tile, j, i)
				self.data[pz + py + x + j / 8] = byte
		else:
			for i in range(0, th):
				py = (y * v.B_TILE_SIZE + i) * self.row_size
				for j in range(0, tw):
					self.data[pz + py + x * v.B_TILE_SIZE + j] = tile[i * v.B_TILE_SIZE + j]
		print str(self)

class Node:
	def __init__(self):
		self.node_id = ~0
		self.name = ""

class NodeBitmap(Node):
	def __init__(self, node_id):
		self.node_id = node_id;
		self.layers = []
		self.width  = 0
		self.height = 0
		self.depth  = 0

	def resize(self, width, height, depth):
		self.width  = width
		self.height = height
		self.depth  = depth
		for l in self.layers:
			l.resize()

	def layer_create(self, layer_id, name, type):
		for i in range(len(self.layers), layer_id + 1):
			self.layers.append(None)
		self.layers[layer_id] = BitmapLayer(self, name, type)
		self.layers[layer_id].resize()

	def layer_lookup(self, layer_id):
		if layer_id < len(self.layers):
			return self.layers[layer_id]
		return None

class NodeDb:
	def __init__(self):
		self.nodes = {}

	def store(self, node):
		self.nodes[node.node_id] = node

	def new(self, node_id, type):
		n = None
		if type == v.BITMAP:
			n = NodeBitmap(node_id)
			v.send_node_subscribe(node_id)
		if n != None:
			self.store(n)
		return n

	def lookup(self, node_id):
		if node_id in self.nodes.keys():
			return self.nodes[node_id]
		return None

# ---------------------------------------------------------------------------------------------

def cb_b_tile_set(node_id, layer_id, tile_x, tile_y, tile_z, type, tile):
	n = the_db.lookup(node_id)
	if n != None:
		l = n.layer_lookup(layer_id)
		if l != None:
			l.tile_set(tile_x, tile_y, tile_z, tile)

def cb_b_layer_create(node_id, layer_id, name, type):
	n = the_db.lookup(node_id)
	if n != None:
		n.layer_create(layer_id, name, type)
		v.send_b_layer_subscribe(node_id, layer_id, 0)

def cb_b_dimensions_set(node_id, width, height, depth):
	n = the_db.lookup(node_id)
	if n != None:
		n.resize(width, height, depth)

def cb_node_name_set(node_id, name):
	n = the_db.lookup(node_id)
	if n != None:
		n.name = name

def cb_node_create(node_id, type, owner):
	global avatar
	if node_id == avatar:
		return
	the_db.new(node_id, type)
	    

def cb_connect_accept(my_avatar, address, host_id):
	global avatar
	avatar = my_avatar
	print "Connected to host " + str(address) + " as avatar " + str(avatar)
	v.send_node_list((1 << v.BITMAP))

print "Running with PyVerse r" + str(v.RELEASE_NUMBER) + "p" + str(v.RELEASE_PATCH)

v.callback_set(v.SEND_CONNECT_ACCEPT,		cb_connect_accept)
v.callback_set(v.SEND_NODE_CREATE,		cb_node_create)
v.callback_set(v.SEND_NODE_NAME_SET,		cb_node_name_set)
v.callback_set(v.SEND_B_DIMENSIONS_SET,		cb_b_dimensions_set)
v.callback_set(v.SEND_B_LAYER_CREATE,		cb_b_layer_create)
v.callback_set(v.SEND_B_TILE_SET,		cb_b_tile_set)

v.send_connect("printpic", "<secret>", "localhost", 0)

the_db = NodeDb()

while 1:
	v.callback_update(10000)
